import click
from click.testing import CliRunner
import subprocess
import pytest

from ..nidm_query import query

def test_query_failing():
    runner = CliRunner()
    res = runner.invoke(query)
    assert res.exit_code != 0
    assert "Missing option" in res.output

# TODO: adding tests that are passing



#!/usr/bin/env python
#**************************************************************************************
#**************************************************************************************
#  nidm_utils.py
#  License: GPL
#**************************************************************************************
#**************************************************************************************
# Date: 11-28-18                 Coded by: David Keator (dbkeator@gmail.com)
# Filename: nidm_utils.py
#
# Program description:  Tools for working with NIDM-Experiment files
#
#**************************************************************************************
# Development environment: Python - PyCharm IDE
#
#**************************************************************************************
# System requirements:  Python 3.X
# Libraries: pybids, numpy, matplotlib, pandas, scipy, math, dateutil, datetime,argparse,
# os,sys,getopt,csv
#**************************************************************************************
# Start date: 11-28-18
# Update history:
# DATE            MODIFICATION				Who
#
#
#**************************************************************************************
# Programmer comments:
#
#
#**************************************************************************************
#**************************************************************************************

import os,sys
from argparse import ArgumentParser
from rdflib import Graph,util
from rdflib.tools import rdf2dot
from nidm.experiment.Utils import read_nidm
from nidm.experiment.Query import GetParticipantIDs
from nidm.core import Constants
from io import StringIO
from os.path import basename,splitext
import subprocess
from graphviz import Source
import tempfile
import click
from nidm.experiment.tools.click_base import cli

# adding click argument parsing
@cli.command()
@click.option("--nidm_file_list", "-nl", required=True,
              help="A comma separated list of NIDM files with full path")
@click.option("--s", "-s", required=False,is_flag=True,
              help="If parameter set then files will be merged by ndar:src_subjec_id of prov:agents")
@click.option("--out_file", "-o",  required=True,
              help="File to write concatenated NIDM files")


def merge(nidm_file_list, s,out_file):
    """
    This function will merge NIDM files.  See command line parameters for supported merge operations.
    """

    graph = Graph()
    for nidm_file in nidm_file_list.split(','):
        graph.parse(nidm_file,format=util.guess_format(nidm_file))

    # create empty graph
    graph=Graph()
    # start with the first NIDM file and merge the rest into the first
    first=True
    for nidm_file in nidm_file_list.split(','):
        if first:
            graph.parse(nidm_file,format=util.guess_format(nidm_file))
            first=False
        # if argument -s is set then merge by subject IDs
        elif s:
            # first get all subject UUIDs in current nidm_file
            subj = GetParticipantIDs([nidm_file])

            # for each UUID / subject ID look in graph and see if you can find the same ID.  If so get the UUID of
            # that prov:agent and change all the UUIDs in nidm_file to match then concatenate the two graphs.
            query = '''

                PREFIX prov:<http://www.w3.org/ns/prov#>
                PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
                PREFIX ndar: <https://ndar.nih.gov/api/datadictionary/v2/dataelement/>
                PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                PREFIX prov:<http://www.w3.org/ns/prov#>

                SELECT DISTINCT ?uuid ?ID
                WHERE {

                        ?uuid a prov:Person ;
                            %s ?ID .
                FILTER(?ID =
                ''' % Constants.NIDM_SUBJECTID

            first = True
            for ID in subj['ID']:
                if first:
                    query = query + "\"" + ID + "\""
                    first = False
                else:
                    query = query + "|| ?ID= \"" + ID + "\""

            query = query + ") }"

            qres = graph.query(query)

            # if len(qres) > 0 then we have matches so load the nidm_file into a temporary graph so we can
            # make changes to it then concatenate it.
            if len(qres) > 0:
                tmp = Graph()
                tmp.parse(nidm_file,format=util.guess_format(nidm_file))

                # for each ID in the merged graph that matches an ID in the nidm_file graph
                for row in qres:
                    # find the UUID in the subj data frame for the matching ID and change all triples that reference
                    # this uuid to the one in row['uuid']
                    uuid_to_replace = (subj[subj['ID'].str.match(row['ID'])])['uuid'].values[0]

                    for s,p,o in tmp.triples((None,None,None)):
                        if (s == uuid_to_replace):
                            #print("replacing subject in triple %s %s %s with %s" %(s,p,o,uuid_to_replace))
                            tmp.set((uuid_to_replace,p,o))
                        elif (o == uuid_to_replace):
                            #print("replacing object in triple %s %s %s with %s" %(s,p,o,uuid_to_replace))
                            tmp.set((s,p,uuid_to_replace))
                        elif (p == uuid_to_replace):
                            #print("replacing predicate in triple %s %s %s with %s" %(s,p,o,uuid_to_replace))
                            tmp.set((s,uuid_to_replace,o))

            # merge updated graph
            graph = graph + tmp

    graph.serialize(out_file, format='turtle')



if __name__ == "__main__":
   merge()


"""Python implementation of the ProvONE Model Specification

References:

ProvONE: http://vcvcomputing.com/provone/provone.html
"""

import logging
import os
import io
import shutil
import tempfile
from six.moves.urllib.parse import urlparse

import nidm.core.serializers
from prov.constants import PROV_N_MAP, PROV_ATTR_STARTTIME, PROV_ATTR_ENDTIME, \
	PROV_ATTR_TIME, PROV_DERIVATION, PROV_GENERATION, PROV_USAGE, \
	PROV_COMMUNICATION, PROV_ASSOCIATION, PROV_ATTRIBUTION, PROV_ATTR_COLLECTION, \
	PROV_ATTR_USAGE, PROV_MEMBERSHIP
from prov.model import ProvEntity, ProvAgent, ProvDocument, ProvAttribution, \
	PROV_REC_CLS, ProvActivity, _ensure_datetime, ProvAssociation, \
	ProvCommunication, ProvDerivation, ProvRelation, ProvGeneration, ProvUsage, \
	ProvMembership
from .Constants import PROVONE_N_MAP, PROVONE_PROCESS, PROVONE_INPUTPORT, \
	PROVONE_OUTPUTPORT, PROVONE_DATA, PROVONE_DATALINK, PROVONE_SEQCTRLLINK, \
	PROVONE_USER, PROVONE_PROCESSEXEC, PROVONE_ATTR_PROCESS, PROVONE_ATTR_USER, \
	PROVONE_ATTR_PROCESSEXEC, PROVONE_ATTR_PLAN, \
	PROVONE_ATTR_INFORMED, PROVONE_ATTR_INFORMANT, \
	PROVONE_ATTR_GENERATED_DATA, PROVONE_ATTR_USED_DATA, \
	PROVONE_ATTR_GENERATION, \
	PROVONE_ATTR_DATA, PROVONE_ATTR_INPUTPORT, \
	PROVONE_HASINPORT, PROVONE_ATTR_OUTPUTPORT, PROVONE_HASOUTPORT, \
	PROVONE_HASSUBPROCESS, PROVONE_ATTR_DATALINK, PROVONE_INPORTTODL, \
	PROVONE_OUTPORTTODL, PROVONE_DLTOOUTPORT, PROVONE_DLTOINPORT, \
	PROVONE_ATTR_SEQCTRLLINK, PROVONE_CLTODESTP, PROVONE_SOURCEPTOCL, \
	PROVONE_DATAONLINK, PROVONE_HASDEFAULTPARAM, \
	PROVONE_ATTR_GENERATED_PROCESS, PROVONE_ATTR_USED_PROCESS, PROVONE_ISPARTOF, \
	PROVONE_ATTR_RELATED_PREXEC, PROVONE_ATTR_USED_PREXEC, \
	PROVONE_ATTR_CHILD_PREXEC, PROVONE_MEMBERSHIP

__author__ = 'Sanu Ann Abraham'
__email__ = 'sanuann@mit.edu'

logger = logging.getLogger(__name__)

# add ProvOne Notation mapping to Prov_N_MAP dict
#PROV_N_MAP.update(PROVONE_N_MAP)
PROVONE_N_MAP.update(PROV_N_MAP)


class ProvPlan(ProvEntity):
	"""
	ProvONE Plan element
	"""
	pass


class Process(ProvEntity):
	"""
	ProvONE Process element """

	_prov_type = PROVONE_PROCESS


class InputPort(ProvEntity):
	""" ProvONE Input Port element """

	_prov_type = PROVONE_INPUTPORT


class OutputPort(ProvEntity):
	""" ProvONE Output Port element"""

	_prov_type = PROVONE_OUTPUTPORT


class Data(ProvEntity):
	"""
	basic unit of information consumed or produced by a Process. Multiple Data items may be grouped into a Collection.
	"""
	_prov_type = PROVONE_DATA


class DataLink(ProvEntity):
	""" ProvONE DataLink Element """

	_prov_type = PROVONE_DATALINK


class SeqCtrlLink(ProvEntity):
	""" ProvONE SeqCtrlLink Element """

	_prov_type = PROVONE_SEQCTRLLINK


class User(ProvAgent):
	"""ProvONE User element."""

	_prov_type = PROVONE_USER


class ProcessExec(ProvActivity):
	""" ProvONE Process Execution element. """

	_prov_type = PROVONE_PROCESSEXEC


class Attribution(ProvAttribution):
	"""ProvONE Attribution relationship."""

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS, PROVONE_ATTR_USER)

	_prov_type = PROV_ATTRIBUTION


class Association(ProvAssociation):
	"""Provenance Association relationship."""
	FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESSEXEC, PROVONE_ATTR_PROCESS,
						 PROVONE_ATTR_PLAN)

	_prov_type = PROV_ASSOCIATION


class Communication(ProvCommunication):
	"""Provenance Communication relationship."""

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_INFORMED, PROVONE_ATTR_INFORMANT)

	_prov_type = PROV_COMMUNICATION


class Derivation(ProvDerivation):
    """Provenance Derivation relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_GENERATED_DATA, PROVONE_ATTR_USED_DATA,
						 PROVONE_ATTR_PROCESSEXEC, PROVONE_ATTR_GENERATION,
						 PROV_ATTR_USAGE)

    _prov_type = PROV_DERIVATION


class Generation(ProvGeneration):
    """Provenance Generation relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATA, PROVONE_ATTR_PROCESSEXEC, PROV_ATTR_TIME)

    #_prov_type = PROV_GENERATION


class Usage(ProvUsage):
    """Provenance Usage relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESSEXEC, PROVONE_ATTR_DATA, PROV_ATTR_TIME)

    #_prov_type = PROV_USAGE


class Partnership(ProvRelation):
    """Provenance Membership relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_USED_PREXEC,
						 PROVONE_ATTR_CHILD_PREXEC)

    _prov_type = PROVONE_ISPARTOF


# Component 6: Collections
class Membership(ProvMembership):
    """Provenance Membership relationship."""

    FORMAL_ATTRIBUTES = (PROV_ATTR_COLLECTION, PROVONE_ATTR_DATA)

    #_prov_type = PROV_MEMBERSHIP


class HasInput(ProvRelation):
	"""ProvONE HasInput Port relationship."""

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS,
						 PROVONE_ATTR_INPUTPORT)

	_prov_type = PROVONE_HASINPORT


class HasOutput(ProvRelation):
	"""ProvONE HasOutput Port relationship."""

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS, PROVONE_ATTR_OUTPUTPORT)

	_prov_type = PROVONE_HASOUTPORT


class HasSubProcess(ProvRelation):
	"""ProvONE Has SubProcess relationship."""

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_USED_PROCESS, PROVONE_ATTR_GENERATED_PROCESS)

	_prov_type = PROVONE_HASSUBPROCESS


class InToDL(ProvRelation):
	""" ProvONE InPort to DL relationship """

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_INPUTPORT, PROVONE_ATTR_DATALINK)

	_prov_type = PROVONE_INPORTTODL


class OutToDL(ProvRelation):
	""" ProvONE Output port to DL relationship """

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_OUTPUTPORT, PROVONE_ATTR_DATALINK)

	_prov_type = PROVONE_OUTPORTTODL


class DLtoOutPort(ProvRelation):
	""" ProvONE DL to Output port relationship """

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATALINK, PROVONE_ATTR_OUTPUTPORT)

	_prov_type = PROVONE_DLTOOUTPORT


class DLtoInPort(ProvRelation):
	""" ProvONE DL to Input port relationship """

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATALINK, PROVONE_ATTR_INPUTPORT)

	_prov_type = PROVONE_DLTOINPORT


class CLtoDestP(ProvRelation):
	"""ProvONE CLtoDestP relationship."""

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_SEQCTRLLINK, PROVONE_ATTR_PROCESS)

	_prov_type = PROVONE_CLTODESTP


class SourcePtoCL(ProvRelation):
	"""ProvONE SourcePtoCL relationship."""

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS, PROVONE_ATTR_SEQCTRLLINK)

	_prov_type = PROVONE_SOURCEPTOCL


class DataLinkage(ProvRelation):
	""" ProvONE dataOnLink relationship """

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATA, PROVONE_ATTR_DATALINK,
						 PROVONE_ATTR_PROCESS)

	_prov_type = PROVONE_DATAONLINK


class Parameterization(ProvRelation):
	""" ProvONE hasDefaultParam relationship. """

	FORMAL_ATTRIBUTES = (PROVONE_ATTR_INPUTPORT, PROVONE_ATTR_DATA)

	_prov_type = PROVONE_HASDEFAULTPARAM


class Workflow(Process, ):
	pass


#  Class mappings from PROVONE record type
PROV_REC_CLS.update({
	PROVONE_PROCESS: Process,
	PROVONE_PROCESSEXEC: ProcessExec,
	PROVONE_DATA: Data,
	PROV_ATTRIBUTION: Attribution,
	PROV_ASSOCIATION: Association,
	PROV_COMMUNICATION:  Communication,
	PROV_DERIVATION: Derivation,
	PROV_GENERATION: Generation,
	PROV_USAGE: Usage,
	PROVONE_INPUTPORT: InputPort,
	PROVONE_HASINPORT: HasInput,
	PROVONE_OUTPUTPORT: OutputPort,
	PROVONE_HASOUTPORT: HasOutput,
	PROVONE_HASSUBPROCESS: HasSubProcess,
	PROVONE_DATALINK: DataLink,
	PROVONE_INPORTTODL: InToDL,
	PROVONE_SEQCTRLLINK: SeqCtrlLink,
	PROVONE_CLTODESTP: CLtoDestP,
	PROVONE_SOURCEPTOCL: SourcePtoCL,
	PROVONE_OUTPORTTODL: OutToDL,
	PROVONE_DLTOOUTPORT: DLtoOutPort,
	PROVONE_DLTOINPORT: DLtoInPort,
	PROVONE_DATAONLINK: DataLinkage,
	PROVONE_HASDEFAULTPARAM: Parameterization,
	PROVONE_USER: User,
	PROVONE_ISPARTOF: Partnership,
	PROV_MEMBERSHIP: Membership,

})


class ProvONEDocument(ProvDocument):
	""" ProvONE Document"""

	def __repr__(self):
		return '<ProvONEDocument>'

	def process(self, identifier, other_attributes=None):
		"""
		Creates a new process.

		:param identifier: Identifier for new process.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(PROVONE_PROCESS, identifier, None,
							   other_attributes)

	def user(self, identifier, other_attributes=None):
		"""
		Creates a new user.

		:param identifier: Identifier for new user.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(PROVONE_USER, identifier, None, other_attributes)

	def data(self, identifier, other_attributes=None):
		"""
		Creates a new data.

		:param identifier: Identifier for new data.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(PROVONE_DATA, identifier, None,
							   other_attributes)

	def attribution(self, process_spec, user, identifier=None,
					other_attributes=None):
		"""
		Creates a new attribution record between a process specification and an user.

		:param process_spec: ProcessSpecification or a string identifier for the process spec (relationship
			source).
		:param user: User or string identifier of the user involved in the
			attribution (relationship destination).
		:param identifier: Identifier for new attribution record.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROV_ATTRIBUTION, identifier, {
				PROVONE_ATTR_PROCESS: process_spec,
				PROVONE_ATTR_USER: user
			},
			other_attributes
		)

	def processExec(self, identifier, startTime=None, endTime=None,
					other_attributes=None):
		"""
		Creates a new process execution.

		:param identifier: Identifier for new process execution.
		:param startTime: Optional start time for the process execution (default:
		None).
			Either a :py:class:`datetime.datetime` object or a string that can be
			parsed by :py:func:`dateutil.parser`.
		:param endTime: Optional end time for the process execution (default: None).
			Either a :py:class:`datetime.datetime` object or a string that can be
			parsed by :py:func:`dateutil.parser`.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROVONE_PROCESSEXEC, identifier, {
				PROV_ATTR_STARTTIME: _ensure_datetime(startTime),
				PROV_ATTR_ENDTIME: _ensure_datetime(endTime)
			},
			other_attributes
		)

	def association(self, process_exec, process_spec=None, plan=None,
					identifier=None, other_attributes=None):
		"""
		Creates a new association record for a process execution.

		:param process_exec: Process Execution or a string identifier for the
			process execution.
		:param process_spec: Process Spec or string identifier of the process
			involved in the association (default: None).
		:param plan: Optionally extra entity to state qualified association through
			an internal plan (default: None).
		:param identifier: Identifier for new association record.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROV_ASSOCIATION, identifier, {
				PROVONE_ATTR_PROCESSEXEC: process_exec,
				PROVONE_ATTR_PROCESS: process_spec,
				PROVONE_ATTR_PLAN: plan
			},
			other_attributes
		)

	def derivation(self, generatedData, usedData, process_exec=None,
				   generation=None, usage=None,
				   identifier=None, other_attributes=None):
		"""
		Creates a new derivation record for a generated data from a used data.

		:param generatedData: Data or a string identifier for the generated
			data (relationship source).
		:param usedData: Data or a string identifier for the used data
			(relationship destination).
		:param process_exec: Process execution or string identifier of the
			processExec involved in the derivation (default: None).
		:param generation: Optionally extra activity to state qualified generation
			through a generation (default: None).
		:param usage: XXX (default: None).
		:param identifier: Identifier for new derivation record.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		attributes = {PROVONE_ATTR_GENERATED_DATA: generatedData,
					  PROVONE_ATTR_USED_DATA: usedData,
					  PROVONE_ATTR_PROCESSEXEC: process_exec,
					  PROVONE_ATTR_GENERATION: generation,
					  PROV_ATTR_USAGE: usage}
		return self.new_record(
			PROV_DERIVATION, identifier, attributes, other_attributes
		)

	def generation(self, data, process_exec=None, time=None, identifier=None,
				   other_attributes=None):
		"""
		Creates a new generation record for a data.

		:param data: Data or a string identifier for the data.
		:param process_exec: Process execution or string identifier of the
		process_exec involved in the generation (default: None).
		:param time: Optional time for the generation (default: None).
			Either a :py:class:`datetime.datetime` object or a string that can be
			parsed by :py:func:`dateutil.parser`.
		:param identifier: Identifier for new generation record.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROV_GENERATION, identifier, {
				PROVONE_ATTR_DATA: data,
				PROVONE_ATTR_PROCESSEXEC: process_exec,
				PROV_ATTR_TIME: _ensure_datetime(time)
			},
			other_attributes
		)

	def usage(self, process_exec, data=None, time=None, identifier=None,
			  other_attributes=None):
		"""
		Creates a new usage record for a process execution.

		:param process_exec: Process Execution or a string identifier for the
			processExec.
		:param data: Data or string identifier of the data involved in
			the usage relationship (default: None).
		:param time: Optional time for the usage (default: None).
			Either a :py:class:`datetime.datetime` object or a string that can be
			parsed by :py:func:`dateutil.parser`.
		:param identifier: Identifier for new usage record.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROV_USAGE, identifier, {
				PROVONE_ATTR_PROCESSEXEC: process_exec,
				PROVONE_ATTR_DATA: data,
				PROV_ATTR_TIME: _ensure_datetime(time)},
			other_attributes
		)

	def communication(self, informed, informant, identifier=None,
					  other_attributes=None):
		"""
		Creates a new communication record for a process exec.

		:param informed: The informed processExec (relationship destination).
		:param informant: The informing processExec (relationship source).
		:param identifier: Identifier for new communication record.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROV_COMMUNICATION, identifier, {
				PROVONE_ATTR_INFORMED: informed,
				PROVONE_ATTR_INFORMANT: informant
			},
			other_attributes
		)

	def input_port(self, identifier, other_attributes=None):
		"""
		Creates a new input port.

		:param identifier: Identifier for new input port.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(PROVONE_INPUTPORT, identifier, None,
							   other_attributes)

	def has_in_ports(self, process, in_ports, identifier=None,
					 other_attributes=None):
		"""
		Creates a new input port record for a process.

		:param process: Process or a string identifier for the
			process(relationship source).
		:param in_ports: Input Port or string identifier for the used input port (
			relationship destination).
		:param identifier: Identifier for new input port membership.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROVONE_HASINPORT, identifier, {
			PROVONE_ATTR_PROCESS: process,
			PROVONE_ATTR_INPUTPORT: in_ports,
			},
			other_attributes
		)

	def output_port(self, identifier, other_attributes=None):
		"""
		Creates a new output port.

		:param identifier: Identifier for new output port.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(PROVONE_OUTPUTPORT, identifier, None,
							   other_attributes)

	def has_out_ports(self, process, out_ports, identifier=None,
					  other_attributes=None):
		"""
		Creates a new input port record for a process.

		:param process: Process or a string identifier for the
			process(relationship source).
		:param out_ports: Output Port or string identifier for the used output
		port (relationship destination).
		:param identifier: Identifier for new output port membership.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROVONE_HASOUTPORT, identifier, {
			PROVONE_ATTR_PROCESS: process,
			PROVONE_ATTR_OUTPUTPORT: out_ports,
			},
			other_attributes
		)

	def has_sub_process(self, used_process, generated_process, identifier=None,
						other_attributes=None ):
		"""
				Creates a new has-sub-process record for a generated process from a
				used process.

				:param used_process: Process or a string identifier for the
				used process (relationship source).
				:param generated_process: Process or a string identifier for the
				generated process (relationship destination).
				:param identifier: Identifier for new sub-process record.
				:param other_attributes: Optional other attributes as a dictionary or list
					of tuples to be added to the record optionally (default: None).
				"""
		attributes = {PROVONE_ATTR_USED_PROCESS: used_process,
					  PROVONE_ATTR_GENERATED_PROCESS: generated_process}
		return self.new_record(
			PROVONE_HASSUBPROCESS, identifier, attributes, other_attributes
		)

	def dataLink(self, identifier, other_attributes=None):
		"""
		Creates a new data link.

		:param identifier: Identifier for new data link.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(PROVONE_DATALINK, identifier, None,
							   other_attributes)

	def inPortToDL(self, in_port, dt_link, identifier=None,
					  other_attributes=None):
		"""

		:param in_port: Input port or a string identifier for the
			in_port(relationship source).
		:param dt_link: Data Link or string identifier for the used data link (
			relationship destination).
		:param identifier: Identifier for new data link membership.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(
			PROVONE_INPORTTODL, identifier, {
				PROVONE_ATTR_INPUTPORT: in_port,
				PROVONE_ATTR_DATALINK: dt_link,
			},
			other_attributes
		)

	def outPortToDL(self, out_port, dt_link, identifier=None,
					  other_attributes=None):
		"""

		:param out_port: Output port or a string identifier for the
			out_port(relationship source).
		:param dt_link: Data Link or string identifier for the used data link (
			relationship destination).
		:param identifier: Identifier for new output-data link membership.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(
			PROVONE_OUTPORTTODL, identifier, {
				PROVONE_ATTR_OUTPUTPORT: out_port,
				PROVONE_ATTR_DATALINK: dt_link,
			},
			other_attributes
		)

	def DLToOutPort(self, dt_link, out_port, identifier=None,
					  other_attributes=None):
		"""

		:param dt_link: Data Link or string identifier for the used data link (
			relationship source).
		:param out_port: Output port or a string identifier for the
			out_port(relationship destination).
		:param identifier: Identifier for new data link-output membership.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(
			PROVONE_DLTOOUTPORT, identifier, {
				PROVONE_ATTR_DATALINK: dt_link,
				PROVONE_ATTR_OUTPUTPORT: out_port,
			},
			other_attributes
		)

	def DLToInPort(self, dt_link, in_port, identifier=None,
					  other_attributes=None):
		"""

		:param dt_link: Data Link or string identifier for the used data link (
			relationship source).
		:param in_port: Input port or a string identifier for the in_port (
			relationship destination).
		:param identifier: Identifier for new data link-output membership.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(
			PROVONE_DLTOINPORT, identifier, {
				PROVONE_ATTR_DATALINK: dt_link,
				PROVONE_ATTR_INPUTPORT: in_port,
			},
			other_attributes
		)

	def seqCtrlLink(self, identifier, other_attributes=None):
		"""
		Creates a new seq ctrl link.

		:param identifier: Identifier for new seq ctrl link.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(PROVONE_SEQCTRLLINK, identifier, None,
							   other_attributes)

	def control_link_to_process(self, used_cntrl_link, used_process,
								identifier=None, other_attributes=None):
		"""

		:param used_cntrl_link: Control Link or a string identifier for the used
		control link (relationship source).
		:param used_process: Data Link or string identifier for the used process
		(relationship destination).
		:param identifier: Identifier for new control link to process relation.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(
			PROVONE_CLTODESTP, identifier, {
				PROVONE_ATTR_SEQCTRLLINK: used_cntrl_link,
				PROVONE_ATTR_PROCESS: used_process,
			},
			other_attributes)

	def process_to_control_link(self, used_process, used_cntrl_link,
								identifier=None, other_attributes=None):
		"""

		:param used_process: Process or string identifier for the used process
		(relationship source).
		:param used_cntrl_link: Control Link or a string identifier for the used
		control link (relationship destination).
		:param identifier: Identifier for new process to control link relation.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(
			PROVONE_SOURCEPTOCL, identifier, {
				PROVONE_ATTR_PROCESS: used_process,
				PROVONE_ATTR_SEQCTRLLINK: used_cntrl_link,
			},
			other_attributes)

	def linkage(self, data_item, dl_link, related_process=None,
					 identifier=None,
				   other_attributes=None):
		"""

		:param data_item: Data or string identifier for the associated data (
		relationship source).
		:param dl_link: Data link or string identifier for the data link (
		relationship destination).
		:param related_process: Process or string identifier of the data link (default=None)
		:param identifier: Identifier for new data-on-link relation.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).

		"""
		return self.new_record(
			PROVONE_DATAONLINK, identifier, {
				PROVONE_ATTR_DATA: data_item,
				PROVONE_ATTR_DATALINK: dl_link,
				PROVONE_ATTR_PROCESS: related_process,
			},
			other_attributes)

	def parameterization(self, in_port, data_item, identifier=None,
						 other_attributes=None):
		"""

		:param in_port: InputPort or string identifier for the associated in port (
		relationship source).
		:param data_item: Data item or string identifier for the associated
		default parameter (relationship destination).
		:param identifier: Identifier for new default parameter relation.
		:param other_attributes: Optional other attributes as a dictionary or
			list of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROVONE_HASDEFAULTPARAM, identifier, {
				PROVONE_ATTR_INPUTPORT: in_port,
				PROVONE_ATTR_DATA: data_item,
			},
			other_attributes)

	def serialize(self, destination=None, format='json', **args):
		"""
		Serialize the :py:class:`ProvDocument` to the destination.

		Available serializers can be queried by the value of
		`:py:attr:~prov.serializers.Registry.serializers` after loading them via
		`:py:func:~prov.serializers.Registry.load_serializers()`.

		:param destination: Stream object to serialize the output to. Default is
			`None`, which serializes as a string.
		:param format: Serialization format (default: 'json'), defaulting to
			PROV-JSON.
		:return: Serialization in a string if no destination was given,
			None otherwise.
		"""
		serializer = nidm.core.serializers.get(format)(self)
		if destination is None:
			stream = io.StringIO()
			serializer.serialize(stream, **args)
			return stream.getvalue()
		if hasattr(destination, "write"):
			stream = destination
			serializer.serialize(stream, **args)
		else:
			location = destination
			scheme, netloc, path, params, _query, fragment = urlparse(location)
			if netloc != "":
				print("WARNING: not saving as location " +
					  "is not a local file reference")
				return
			fd, name = tempfile.mkstemp()
			stream = os.fdopen(fd, "wb")
			serializer.serialize(stream, **args)
			stream.close()
			if hasattr(shutil, "move"):
				shutil.move(name, path)
			else:
				shutil.copy(name, path)
				os.remove(name)

	def is_part_of(self, used_processex, child_processex, identifier=None,
					  other_attributes=None):
		"""
		Creates a new is-part-of record for a process exec.

		:param used_processex: The parent processExec (relationship source).
		:param child_processex: The child processExec (relationship destination).
		:param identifier: Identifier for new is-part-of record.
		:param other_attributes: Optional other attributes as a dictionary or list
			of tuples to be added to the record optionally (default: None).
		"""
		return self.new_record(
			PROVONE_ISPARTOF, identifier, {
				PROVONE_ATTR_USED_PREXEC: used_processex,
				PROVONE_ATTR_CHILD_PREXEC: child_processex
			},
			other_attributes
		)

	def membership(self, collection, data):
		"""
		Creates a new membership record for data to a collection.

		:param collection: Collection the data is to be added to.
		:param data: Data to be added to the collection.
		"""
		return self.new_record(
			PROV_MEMBERSHIP, None, {
				PROV_ATTR_COLLECTION: collection,
				PROVONE_ATTR_DATA: data
			}
		)

	# Aliases
	wasAttributedTo = attribution
	wasAssociatedWith = association
	wasDerivedFrom = derivation
	wasGeneratedBy = generation
	wasInformedBy = communication
	used = usage
	hasInPort = has_in_ports
	hasOutPort = has_out_ports
	hasSubProcess = has_sub_process
	CLtoDestP = control_link_to_process
	sourcePToCL = process_to_control_link
	hasDefaultParam = parameterization
	dataOnLink = linkage
	isPartOf = is_part_of
	hadMember = membership



#!/usr/bin/env python
#**************************************************************************************
#**************************************************************************************
#  nidm_utils.py
#  License: GPL
#**************************************************************************************
#**************************************************************************************
# Date: 11-28-18                 Coded by: David Keator (dbkeator@gmail.com)
# Filename: nidm_utils.py
#
# Program description:  Tools for working with NIDM-Experiment files
#
#**************************************************************************************
# Development environment: Python - PyCharm IDE
#
#**************************************************************************************
# System requirements:  Python 3.X
# Libraries: pybids, numpy, matplotlib, pandas, scipy, math, dateutil, datetime,argparse,
# os,sys,getopt,csv
#**************************************************************************************
# Start date: 11-28-18
# Update history:
# DATE            MODIFICATION				Who
#
#
#**************************************************************************************
# Programmer comments:
#
#
#**************************************************************************************
#**************************************************************************************

import os,sys
from argparse import ArgumentParser
from rdflib import Graph,util
from rdflib.tools import rdf2dot
from nidm.experiment.Utils import read_nidm
from io import StringIO
from os.path import basename,splitext
import subprocess
from graphviz import Source
import tempfile
import click
from nidm.experiment.tools.click_base import cli

# adding click argument parsing
@cli.command()
@click.option("--nidm_file_list", "-nl", required=True,
              help="A comma separated list of NIDM files with full path")



def visualize(nidm_file_list):
    '''
    This command will produce a visualization(pdf) of the supplied NIDM files named the same as the input files and
    stored in the same directories.
    '''

    for nidm_file in nidm_file_list.split(','):
        # read in nidm file
        project=read_nidm(nidm_file)

        # split path and filename for output file writing
        file_parts = os.path.split(nidm_file)

        # write graph as nidm filename + .pdf
        project.save_DotGraph(filename=os.path.join(file_parts[0], os.path.splitext(file_parts[1])[0] + '.pdf'), format='pdf' )



if __name__ == "__main__":
   visualize()



import urllib
import re

import pytest
import rdflib

from nidm.experiment import Project, Session, AssessmentAcquisition, AssessmentObject, Acquisition, AcquisitionObject, Query
from nidm.core import Constants
from nidm.experiment.tools.rest import RestParser
import os
from pathlib import Path
from rdflib import Graph, util, URIRef
import json


from prov.model import ProvAgent


REST_TEST_FILE = './agent.ttl'
BRAIN_VOL_FILES = ['./cmu_a.nidm.ttl', './caltech.nidm.ttl']
test_person_uuid = ""
test_p2_subject_uuids = []
stat_test_project_uuid = None
restParser = RestParser(verbosity_level=0, output_format=RestParser.OBJECT_FORMAT)

@pytest.fixture(scope="module", autouse="True")
def setup():

    for f in ['./cmu_a.nidm.ttl', 'caltech.nidm.ttl']:
        if Path(f).is_file():
            os.remove(f)

    if not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )

    if not Path('./caltech.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/Caltech/nidm.ttl",
            "caltech.nidm.ttl"
        )

def getSampleProjectUUID():
    global stat_test_project_uuid
    if stat_test_project_uuid == None:
        projects = restParser.run(['./cmu_a.nidm.ttl'], '/projects')
        cmu_test_project_uuid = projects[0]
    return cmu_test_project_uuid


def test_project_statistics():
    AGE_CUTOFF = 30

    project = getSampleProjectUUID()

    # basics stats
    basic_project_stats = restParser.run(BRAIN_VOL_FILES, "/statistics/projects/{}".format(project))
    assert 'title' in basic_project_stats

    # basics stats with subjects
    project_stats_with_subjects = restParser.run(BRAIN_VOL_FILES, "/statistics/projects/{}?fields=subjects".format(project))
    assert 'title' in project_stats_with_subjects
    assert 'subjects' in project_stats_with_subjects

    # filtered subjects stats
    filtered_stats_with_subjects = restParser.run(BRAIN_VOL_FILES, "/statistics/projects/{}?fields=subjects&filter=instruments.AGE_AT_SCAN gt {}".format(project, AGE_CUTOFF))
    assert 'title' in filtered_stats_with_subjects
    assert 'subjects' in filtered_stats_with_subjects
    assert len(filtered_stats_with_subjects['subjects']) < len(project_stats_with_subjects['subjects'])


    # filtered subjects instrument stats
    age_stats = restParser.run(BRAIN_VOL_FILES, "/statistics/projects/{}?fields=instruments.AGE_AT_SCAN&filter=instruments.AGE_AT_SCAN gt {}".format(project, AGE_CUTOFF))
    assert 'title' in age_stats
    assert 'subjects' in age_stats
    assert len(age_stats['subjects']) < len(project_stats_with_subjects['subjects'])
    assert 'AGE_AT_SCAN' in age_stats
    for x in ['max', 'min', 'mean', 'median', 'standard_deviation']:
        assert x in age_stats['AGE_AT_SCAN']
    assert age_stats['AGE_AT_SCAN']['min'] > AGE_CUTOFF
    assert age_stats['AGE_AT_SCAN']['median'] >= age_stats['AGE_AT_SCAN']['min']
    assert age_stats['AGE_AT_SCAN']['median'] <= age_stats['AGE_AT_SCAN']['max']

    # filtered subjects instrument and derivative stats
    derivative_stats = restParser.run(BRAIN_VOL_FILES, "/statistics/projects/{}?fields=instruments.AGE_AT_SCAN,derivatives.Right-Hippocampus (mm^3)&filter=instruments.AGE_AT_SCAN gt {}".format(project, AGE_CUTOFF))
    assert 'title' in derivative_stats
    assert 'subjects' in derivative_stats
    assert len(derivative_stats['subjects']) < len(project_stats_with_subjects['subjects'])
    for field in ['Right-Hippocampus (mm^3)', 'AGE_AT_SCAN']:
        assert field in derivative_stats
        for x in ['max', 'min', 'mean', 'median', 'standard_deviation']:
            assert x in derivative_stats[field]

def test_project_statistics_fields():

    project = getSampleProjectUUID()

    # ask for a field based on URI tail
    derivative_stats = restParser.run(BRAIN_VOL_FILES, "/statistics/projects/{}?fields=derivatives.fsl_000020".format(project))
    assert 'title' in derivative_stats
    assert 'subjects' in derivative_stats
    assert len(derivative_stats['subjects']) > 0
    for field in ['fsl_000020']:
        assert field in derivative_stats
        for x in ['max', 'min', 'mean', 'median', 'standard_deviation']:
            assert x in derivative_stats[field]

    # ask for a field based on URI tail
    derivative_stats = restParser.run(BRAIN_VOL_FILES, "/statistics/projects/{}?fields=derivatives.fsl_000020,instruments.AGE_AT_SCAN".format(project))
    assert 'title' in derivative_stats
    assert 'subjects' in derivative_stats
    assert len(derivative_stats['subjects']) > 0
    for field in ['fsl_000020', 'AGE_AT_SCAN']:
        assert field in derivative_stats
        for x in ['max', 'min', 'mean', 'median', 'standard_deviation']:
            assert x in derivative_stats[field]


def test_getTailOfURI():
    assert restParser.getTailOfURI('http://purl.org/nidash/fsl#fsl_000020') == 'fsl_000020'
    assert restParser.getTailOfURI('https://surfer.nmr.mgh.harvard.edu/fs_00005') == 'fs_00005'




#!/usr/bin/env python
#**************************************************************************************
#**************************************************************************************
#  nidm_utils.py
#  License: GPL
#**************************************************************************************
#**************************************************************************************
# Date: 11-28-18                 Coded by: David Keator (dbkeator@gmail.com)
# Filename: nidm_utils.py
#
# Program description:  Tools for working with NIDM-Experiment files
#
#**************************************************************************************
# Development environment: Python - PyCharm IDE
#
#**************************************************************************************
# System requirements:  Python 3.X
# Libraries: pybids, numpy, matplotlib, pandas, scipy, math, dateutil, datetime,argparse,
# os,sys,getopt,csv
#**************************************************************************************
# Start date: 11-28-18
# Update history:
# DATE            MODIFICATION				Who
#
#
#**************************************************************************************
# Programmer comments:
#
#
#**************************************************************************************
#**************************************************************************************

import os,sys
from argparse import ArgumentParser
from rdflib import Graph,util
from rdflib.tools import rdf2dot
from nidm.experiment.Utils import read_nidm
from nidm.experiment.Query import GetMergedGraph
from io import StringIO
from os.path import basename,splitext
import subprocess
from graphviz import Source
import tempfile

import click
from nidm.experiment.tools.click_base import cli

# adding click argument parsing
@cli.command()
@click.option("--nidm_file_list", "-nl", required=True,
              help="A comma separated list of NIDM files with full path")
@click.option("--out_file", "-o",  required=True,
              help="File to write concatenated NIDM files")


def concat(nidm_file_list, out_file):
    """
    This function will concatenate NIDM files.  Warning, no merging will be done so you may end up with
    multiple prov:agents with the same subject id if you're concatenating NIDM files from multiple vists of the
    same study.  If you want to merge NIDM files on subject ID see pynidm merge
    """
    #create empty graph
    graph = GetMergedGraph(nidm_file_list.split(','))
    graph.serialize(out_file, format='turtle')



if __name__ == "__main__":
   concat()



import os,sys
import pytest, pdb

from nidm.experiment import Project, Session, Acquisition, AcquisitionObject
from nidm.core import Constants
from nidm.experiment.Utils import load_nidm_owl_files, fuzzy_match_terms_from_graph

def test_loadowl():

    owl_graph = load_nidm_owl_files()
    owl_match = fuzzy_match_terms_from_graph(owl_graph,"WisconsinCardSortingTest")






import urllib
import re

import pytest
import rdflib

from nidm.experiment import Project, Session, AssessmentAcquisition, AssessmentObject, Acquisition, AcquisitionObject, Query
from nidm.core import Constants
from nidm.experiment.tools.rest import RestParser
import os
from pathlib import Path
from rdflib import Graph, util, URIRef
import json


from prov.model import ProvAgent


REST_TEST_FILE = './agent.ttl'
BRAIN_VOL_FILES = ['./cmu_a.nidm.ttl', './caltech.nidm.ttl']
test_person_uuid = ""
test_p2_subject_uuids = []
cmu_test_project_uuid = None
cmu_test_subject_uuid = None

@pytest.fixture(scope="module", autouse="True")
def setup():
    global cmu_test_project_uuid, cmu_test_subject_uuid

    if Path(REST_TEST_FILE).is_file():
        os.remove(REST_TEST_FILE)
    makeTestFile(filename=REST_TEST_FILE, params={'PROJECT_UUID': 'p1', 'PROJECT2_UUID': 'p2'})

    for f in ['./cmu_a.nidm.ttl', 'caltech.nidm.ttl']:
        if Path(f).is_file():
            os.remove(f)

    if not Path('./cmu_a.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
            "cmu_a.nidm.ttl"
        )
        restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
        projects = restParser.run(['./cmu_a.nidm.ttl'], '/projects')
        cmu_test_project_uuid = projects[0]
        subjects = restParser.run(['./cmu_a.nidm.ttl'], '/projects/{}/subjects'.format(cmu_test_project_uuid))
        cmu_test_subject_uuid = subjects[0]

    if not Path('./caltech.nidm.ttl').is_file():
        urllib.request.urlretrieve (
            "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/Caltech/nidm.ttl",
            "caltech.nidm.ttl"
        )



def addData(acq, data):
    acq_entity = AssessmentObject(acquisition=acq)
    for key in data:
        acq_entity.add_attributes({key:data[key]})
    return acq

def makeTestFile(filename, params):
    global test_person_uuid, test_p2_subject_uuids

    nidm_project_name = params.get('NIDM_PROJECT_NAME', False) or "Project_name_sample"
    nidm_project_identifier = params.get('NIDM_PROJECT_IDENTIFIER', False) or 9610
    nidm_project2_identifier = params.get('NIDM_PROJECT_IDENTIFIER', False) or 550
    nidm_project_description = params.get('NIDM_PROJECT_DESCRIPTION', False) or "1234356 Test investigation"
    project_uuid = params.get('PROJECT_UUID', False) or "_proj1"
    project_uuid2 = params.get('PROJECT2_UUID', False) or "_proj2"
    session_uuid = params.get('SESSION_UUID', False) or "_ses1"
    session_uuid2 = params.get('SESSION2_UUID', False) or "_ses2"
    p1kwargs={Constants.NIDM_PROJECT_NAME:nidm_project_name, Constants.NIDM_PROJECT_IDENTIFIER:nidm_project_identifier, Constants.NIDM_PROJECT_DESCRIPTION:nidm_project_description}
    p2kwargs={Constants.NIDM_PROJECT_NAME:nidm_project_name, Constants.NIDM_PROJECT_IDENTIFIER:nidm_project2_identifier, Constants.NIDM_PROJECT_DESCRIPTION:nidm_project_description}

    project = Project(uuid=project_uuid,attributes=p1kwargs)
    session = Session(uuid=session_uuid,project=project)
    acq = Acquisition(uuid="_acq1",session=session)
    acq2 = Acquisition(uuid="_acq2",session=session)
    acq3 = Acquisition(uuid="_acq2",session=session)

    person=acq.add_person(attributes=({Constants.NIDM_SUBJECTID:"a1_9999"}))
    test_person_uuid = (str(person.identifier)).replace("niiri:", "")


    acq.add_qualified_association(person=person,role=Constants.NIDM_PARTICIPANT)

    person2=acq2.add_person(attributes=({Constants.NIDM_SUBJECTID:"a1_8888"}))
    acq2.add_qualified_association(person=person2,role=Constants.NIDM_PARTICIPANT)
    person3=acq3.add_person(attributes=({Constants.NIDM_SUBJECTID:"a2_7777"}))
    acq2.add_qualified_association(person=person3,role=Constants.NIDM_PARTICIPANT)

    project2 = Project(uuid=project_uuid2,attributes=p2kwargs)
    session2 = Session(uuid=session_uuid2,project=project2)
    acq4 = Acquisition(uuid="_acq3",session=session2)
    acq5 = Acquisition(uuid="_acq4",session=session2)

    person4=acq4.add_person(attributes=({Constants.NIDM_SUBJECTID:"a3_6666"}))
    acq4.add_qualified_association(person=person4,role=Constants.NIDM_PARTICIPANT)
    person5=acq5.add_person(attributes=({Constants.NIDM_SUBJECTID:"a4_5555"}))
    acq5.add_qualified_association(person=person5,role=Constants.NIDM_PARTICIPANT)

    # now add some assessment instrument data
    addData(acq,{Constants.NIDM_AGE:9, Constants.NIDM_HANDEDNESS: "R", Constants.NIDM_DIAGNOSIS: "Anxiety"})
    addData(acq2,{Constants.NIDM_AGE:8, Constants.NIDM_HANDEDNESS: "L", Constants.NIDM_DIAGNOSIS: "ADHD"})
    addData(acq4,{Constants.NIDM_AGE:7, Constants.NIDM_HANDEDNESS: "A", Constants.NIDM_DIAGNOSIS: "Depression"})
    addData(acq5,{Constants.NIDM_AGE:6, Constants.NIDM_HANDEDNESS: "R", Constants.NIDM_DIAGNOSIS: "Depression"})

    test_p2_subject_uuids.append( (str(person4.identifier)).replace("niiri:", "") )
    test_p2_subject_uuids.append( (str(person5.identifier)).replace("niiri:", "") )

    with open("a.ttl",'w') as f:
        f.write(project.graph.serialize(None, format='rdf', rdf_format='ttl'))
    with open("b.ttl",'w') as f:
        f.write(project2.graph.serialize(None, format='rdf', rdf_format='ttl'))

    #create empty graph
    graph=Graph()
    for nidm_file in ("a.ttl", "b.ttl"):
         tmp = Graph()
         graph = graph + tmp.parse(nidm_file,format=util.guess_format(nidm_file))

    graph.serialize(filename, format='turtle')

    os.unlink("a.ttl")
    os.unlink("b.ttl")

    with open(filename, "r") as f:
        x = f.read()

    with open("./agent.ttl", "w") as f:
        f.write(x)

def test_uri_project_list():

    kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseII",Constants.NIDM_PROJECT_IDENTIFIER:9610,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation"}
    project = Project(uuid="_123456",attributes=kwargs)
    #save a turtle file
    with open("uritest.ttl",'w') as f:
        f.write(project.serializeTurtle())

    kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseIII",Constants.NIDM_PROJECT_IDENTIFIER:1200,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation2"}
    project = Project(uuid="_654321",attributes=kwargs)
    #save a turtle file
    with open("uritest2.ttl",'w') as f:
        f.write(project.serializeTurtle())

    restParser = RestParser()
    result = restParser.run(['uritest.ttl', 'uritest2.ttl'], '/projects')


    project_uuids = []

    for uuid in result:
        project_uuids.append(uuid)

    assert type(result) == list
    assert len(project_uuids) == 2
    assert "_123456" in project_uuids
    assert "_654321" in project_uuids

    os.remove("uritest.ttl")
    os.remove("uritest2.ttl")


def test_uri_project_id():

    # try with the real brain volume files
    restParser = RestParser()
    result = restParser.run(BRAIN_VOL_FILES, '/projects')
    project = result[0]
    result = restParser.run(BRAIN_VOL_FILES, '/projects/{}'.format(project))

    assert 'dctypes:title' in result
    assert 'sio:Identifier' in result
    assert 'subjects' in result
    assert len(result['subjects']) > 2
    assert 'data_elements' in result
    assert len(result['data_elements']) > 2



def test_uri_projects_subjects_1():
    global test_p2_subject_uuids

    proj_uuid = 'p2'
    restParser = RestParser()
    result = restParser.run([REST_TEST_FILE], '/projects/{}/subjects'.format(proj_uuid))

    assert type(result) == list
    assert len(result) == 2

    assert test_p2_subject_uuids[0] in result
    assert test_p2_subject_uuids[1] in result

def test_uri_subjects():
    global cmu_test_subject_uuid

    restParser = RestParser()
    restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
    result = restParser.run(BRAIN_VOL_FILES, '/subjects/{}'.format(cmu_test_subject_uuid))

    assert type(result) == dict
    assert 'uuid' in result
    assert 'instruments' in result
    assert 'derivatives' in result

    assert cmu_test_subject_uuid == result['uuid']


def test_uri_projects_subjects_id():
    global test_person_uuid

    restParser = RestParser()
    if cmu_test_project_uuid:
        project = cmu_test_project_uuid
    else:
        result = restParser.run(['./cmu_a.nidm.ttl'], '/projects')
        project = result[0]
    result = restParser.run(['./cmu_a.nidm.ttl'], '/projects/{}/subjects'.format(project))
    subject = result[0]

    uri = '/projects/{}/subjects/{}'.format(project,subject)
    result = restParser.run(['./cmu_a.nidm.ttl'], uri)

    assert type(result) == dict
    assert result['uuid'] == subject
    assert len(result['instruments']) > 2

    instruments = result['instruments'].values()
    all_keys = []
    for i in instruments:
        all_keys += i.keys()
    assert 'AGE_AT_SCAN' in all_keys

    assert len(result['derivatives']) > 0



def test_get_software_agents():
    nidm_file = BRAIN_VOL_FILES[0]
    rdf_graph = Query.OpenGraph(nidm_file)

    agents = Query.getSoftwareAgents(rdf_graph)

    assert len(agents) > 0

    isa = URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')


    count = 0
    for a in agents:
        for s, o, p in rdf_graph.triples( (a, isa, Constants.PROV['Agent']) ):
            count += 1

    assert (count == len(agents))



def test_brain_vols():
    restParser = RestParser()
    if cmu_test_project_uuid:
        project = cmu_test_project_uuid
    else:
        project  = (restParser.run(BRAIN_VOL_FILES, '/projects'))[0]
    subjects = restParser.run(BRAIN_VOL_FILES, '/projects/{}/subjects'.format(project))
    subject = subjects[0]

    data = Query.GetDerivativesDataForSubject(BRAIN_VOL_FILES, None, subject)


    assert(len(data) > 0)
    for key in data:
        assert('StatCollectionType' in data[key])
        assert('URI' in data[key])
        assert('values' in data[key])


def test_GetParticipantDetails():

    import time
    start = time.time()

    restParser = RestParser()
    if cmu_test_project_uuid:
        project = cmu_test_project_uuid
    else:
        projects  = restParser.run(BRAIN_VOL_FILES, '/projects')
        project = projects[0]
    import time
    start = time.time()
    subjects = restParser.run(BRAIN_VOL_FILES, '/projects/{}/subjects'.format(project))
    subject = subjects[0]


    Query.GetParticipantInstrumentData( BRAIN_VOL_FILES, project, subject )


    details = Query.GetParticipantDetails( BRAIN_VOL_FILES, project, subject )

    assert ('uuid' in details)
    assert ('id' in details)
    assert ('activity' in details)
    assert ('instruments' in details)
    assert ('derivatives' in details)

    end = time.time()
    runtime = end - start
    # assert (runtime <  4)



def test_CheckSubjectMatchesFilter():
    restParser = RestParser()
    if cmu_test_project_uuid:
        project = cmu_test_project_uuid
    else:
        projects  = restParser.run(BRAIN_VOL_FILES, '/projects')
        project = projects[0]
    subjects = restParser.run(BRAIN_VOL_FILES, '/projects/{}/subjects'.format(project))
    subject = subjects[0]

    derivatives = Query.GetDerivativesDataForSubject(BRAIN_VOL_FILES, project, subject)

    for skey in derivatives:
        for vkey in derivatives[skey]['values']:
            dt = vkey
            val = derivatives[skey]['values'][vkey]['value']
            if (dt and val):
                break

    # find an actual stat and build a matching filter to make sure our matcher passes it
    filter = "derivatives.{} eq {}".format(dt,val)
    assert Query.CheckSubjectMatchesFilter( BRAIN_VOL_FILES, project, subject, filter)


    instruments = Query.GetParticipantInstrumentData( BRAIN_VOL_FILES, project, subject )
    for (i,inst) in instruments.items():
        if 'AGE_AT_SCAN' in inst:
            age = inst['AGE_AT_SCAN']

    older = str(float(age) + 1)
    younger = str(float(age) - 1)

    assert Query.CheckSubjectMatchesFilter( BRAIN_VOL_FILES, project, subject, "instruments.AGE_AT_SCAN eq {}".format( str(age) ) )
    assert (Query.CheckSubjectMatchesFilter( BRAIN_VOL_FILES, project, subject, "instruments.AGE_AT_SCAN lt {}".format( younger ) ) == False)
    assert (Query.CheckSubjectMatchesFilter( BRAIN_VOL_FILES, project, subject, "instruments.AGE_AT_SCAN gt {}".format( younger) ) == True)
    assert Query.CheckSubjectMatchesFilter( BRAIN_VOL_FILES, project, subject, "instruments.AGE_AT_SCAN lt {}".format( older ) )
    assert (Query.CheckSubjectMatchesFilter( BRAIN_VOL_FILES, project, subject, "instruments.AGE_AT_SCAN gt {}".format( older) ) == False)

    eq__format = "instruments.{} eq '{}'".format('WISC_IV_VOCAB_SCALED', 'nan')
    assert Query.CheckSubjectMatchesFilter(BRAIN_VOL_FILES, project, subject, eq__format)
    eq__format = "instruments.{} eq '{}'".format('WISC_IV_VOCAB_SCALED', 'not a match')
    assert (Query.CheckSubjectMatchesFilter( BRAIN_VOL_FILES, project, subject, eq__format ) == False)

def test_ExtremeFilters():
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    if cmu_test_project_uuid:
        project = cmu_test_project_uuid
    else:
        projects  = restParser.run(BRAIN_VOL_FILES, '/projects')
        project = projects[0]

    details = restParser.run(BRAIN_VOL_FILES, '/projects/{}?filter=AGE_AT_SCAN gt 200'.format(project))
    assert len(details['subjects']) == 0
    assert len(details['data_elements']) > 0

    details = restParser.run(BRAIN_VOL_FILES, '/projects/{}?filter=instruments.AGE_AT_SCAN gt 0'.format(project))
    assert len(details['subjects']) > 0
    assert len(details['data_elements']) > 0



def test_OpenGraph():

    g = Query.OpenGraph(BRAIN_VOL_FILES[0])
    assert isinstance(g, rdflib.graph.Graph)

    # if you call OpenGraph with something that is already a graph, it should send it back
    g2 = Query.OpenGraph(g)
    assert isinstance(g, rdflib.graph.Graph)


def test_CDEs():
    def testrun():
        path = os.path.abspath(__file__)

        dir_parts = path.split('/')
        dir_parts = dir_parts[:-4]

        dir_parts.append("core")
        dir_parts.append("cde_dir")
        dir = "/".join(dir_parts)

        graph = Query.getCDEs([
            "{}/ants_cde.ttl".format(dir),
            "{}/fs_cde.ttl".format(dir)
        ])

        units = graph.objects(subject=Constants.FREESURFER['fs_000002'], predicate=Constants.NIDM['hasUnit'])
        count = 0
        for u in units:
            count += 1
            assert str(u) == 'mm^2'

        assert count == 1

    testrun()
    Query.getCDEs.cache = None # clear the memory cache and try again
    testrun() # run a second time to test disk caching.

def assess_one_col_output(txt_output):
    # print (txt_output)
    lines = txt_output.strip().splitlines()
    assert re.search('UUID', lines[0])
    assert re.search('^-+$', lines[1])
    assert is_uuid(lines[2])

    return lines[2]

def is_uuid(uuid):
    return re.search('^[0-9a-z]+-[0-9a-z]+-[0-9a-z]+-[0-9a-z]+-[0-9a-z]+$', uuid) != None

def test_cli_rest_routes():
    rest_parser = RestParser(verbosity_level=0)
    rest_parser.setOutputFormat(RestParser.CLI_FORMAT)

    #
    # / projects
    #

    project_uuid = assess_one_col_output( rest_parser.run(BRAIN_VOL_FILES, "/projects") )


    #
    # /statistics/projects/{}
    #

    txt_out = rest_parser.run(BRAIN_VOL_FILES, "/statistics/projects/{}".format(project_uuid))
    lines = txt_out.strip().splitlines()
    assert re.search('^-+ +-+$', lines[0])
    lines = lines[1:] # done testing line one, slice it off

    split_lines = [ str.split(x) for x in lines ]
    found_gender = found_age_max = found_age_min = found_title = False
    for split in split_lines:
        if len(split) > 0: # skip blank lines between apendicies
            if re.search('title', split[0]): found_title = True
            if re.search('age_max', split[0]): found_age_max = True
            if re.search('age_min', split[0]): found_age_min = True
            if re.search('gender', split[0]): found_gender = True

    assert found_title
    assert found_age_max
    assert found_age_min
    assert found_gender

    #
    # /projects/{}/subjects
    #

    subject_uuid = assess_one_col_output( rest_parser.run(BRAIN_VOL_FILES, '/projects/{}/subjects'.format(project_uuid))  )

    #
    # /projects/{}/subjects/{}/instruments
    #
    # result should be in 3 sections: summary , derivatives, instruments


    inst_text = rest_parser.run(BRAIN_VOL_FILES, '/projects/{}/subjects/{}/'.format(project_uuid, subject_uuid))
    sections = inst_text.split("\n\n")

    # summary tests
    summary_lines = sections[0].strip().splitlines()[1:-1] # first and last lines should be -----
    summary = dict()
    for l in summary_lines:
        summary[l.split()[0]] = l.split()[1]
    inst_uuid = summary['instruments'].split(',')[0]
    deriv_uuid = summary['derivatives'].split(',')[0]
    assert is_uuid(inst_uuid)
    assert is_uuid(deriv_uuid)

    # derivatives test
    deriv_lines = sections[1].strip().splitlines()
    deriv_headers = deriv_lines[0].split()
    heads = ['Derivative_UUID', 'Measurement', 'Label', 'Value', 'Datumtype']
    for i in range(len(heads)):
        assert re.search(heads[i], deriv_headers[i], re.IGNORECASE)
    d_uuid = deriv_lines[2].split()[0]
    assert is_uuid(d_uuid)
    assert d_uuid in summary['derivatives'].split(',')

    #instruments test
    inst_lines = sections[2].strip().splitlines()
    inst_headers = inst_lines[0].split()
    heads = ['Instrument_UUID', 'Category', 'Value']
    for i in range(len(heads)):
        assert re.search(heads[i], inst_headers[i], re.IGNORECASE)
    i_uuid = inst_lines[2].split()[0]
    assert is_uuid(i_uuid)
    assert i_uuid in summary['instruments'].split(',')


def test_project_fields_deriv():
    rest_parser = RestParser(verbosity_level=0)
    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = 'fs_000003'
    project = rest_parser.run( BRAIN_VOL_FILES, "/projects/{}?fields={}".format(cmu_test_project_uuid, field) )

    assert( 'field_values' in project )
    fv = project['field_values']
    assert( type( fv ) == list )
    fields_used = set( [ i["field"] for i in fv ]  )
    assert field in fields_used

def test_project_fields_instruments():
    rest_parser = RestParser(verbosity_level=0)
    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = 'AGE_AT_SCAN'
    project = rest_parser.run( BRAIN_VOL_FILES, "/projects/{}?fields={}".format(cmu_test_project_uuid, field) )

    assert( 'field_values' in project )
    fv = project['field_values']
    assert( type( fv ) == list )
    fields_used = set( [ i["field"] for i in fv ]  )
    assert field in fields_used


def test_project_fields_not_found():
    # test that things don't break if the field isn't in project
    rest_parser = RestParser(verbosity_level=0)
    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = 'not_real_field'
    project = rest_parser.run( BRAIN_VOL_FILES, "/projects/{}?fields={}".format(cmu_test_project_uuid, field) )

    assert( 'field_values' in project )
    fv = project['field_values']
    assert( type( fv ) == list )
    assert len(fv) == 0






#!/usr/bin/env python
#**************************************************************************************
#**************************************************************************************
#  nidm_utils.py
#  License: GPL
#**************************************************************************************
#**************************************************************************************
# Date: 11-28-18                 Coded by: David Keator (dbkeator@gmail.com)
# Filename: nidm_utils.py
#
# Program description:  Tools for working with NIDM-Experiment files
#
#**************************************************************************************
# Development environment: Python - PyCharm IDE
#
#**************************************************************************************
# System requirements:  Python 3.X
# Libraries: pybids, numpy, matplotlib, pandas, scipy, math, dateutil, datetime,argparse,
# os,sys,getopt,csv
#**************************************************************************************
# Start date: 11-28-18
# Update history:
# DATE            MODIFICATION				Who
#
#
#**************************************************************************************
# Programmer comments:
#
#
#**************************************************************************************
#**************************************************************************************

import os,sys
from argparse import ArgumentParser
from rdflib import Graph,util
from rdflib.tools import rdf2dot
from nidm.experiment.Utils import read_nidm
from nidm.experiment.Query import GetMergedGraph
from io import StringIO
from os.path import basename,splitext
import subprocess
from graphviz import Source
import tempfile

import click
from nidm.experiment.tools.click_base import cli

# adding click argument parsing
@cli.command()
@click.option("--nidm_file_list", "-nl", required=True,
              help="A comma separated list of NIDM files with full path")
@click.option("--type", "-t", required=True,type=click.Choice(['turtle', 'jsonld', 'xml-rdf','n3','trig'], case_sensitive=False),
              help="If parameter set then NIDM file will be exported as JSONLD")



def convert(nidm_file_list, type):
    """
    This function will convert NIDM files to various RDF-supported formats and name then / put them in the same
    place as the input file.
    """

    for nidm_file in nidm_file_list.split(','):
        # WIP: for now we use pynidm for jsonld exports to make more human readable and rdflib for everything
        # else.
        if type == 'jsonld':
            # read in nidm file
            project = read_nidm(nidm_file)
            #write jsonld file with same name
            with open(splitext(nidm_file)[0] + ".json", 'w') as f:
                f.write(project.serializeJSONLD())
        elif type == 'turtle':
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(splitext(nidm_file)[0] + ".ttl", format='turtle')
        elif type == 'xml-rdf':
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(splitext(nidm_file)[0] + ".xml", format='pretty-xml')
        elif type == 'n3':
            graph = Graph()
            graph.parse(nidm_file, format=util.guess_format(nidm_file))
            graph.serialize(splitext(nidm_file)[0] + ".n3", format='n3')
        elif type == 'trig':
            # read in nidm file
            project = read_nidm(nidm_file)
            with open(splitext(nidm_file)[0] + ".trig", 'w') as f:
                f.write(project.serializeTrig())
        else:
            print("Error, type is not supported at this time")


if __name__ == "__main__":
   convert()



#!/usr/bin/env python
#**************************************************************************************
#**************************************************************************************
#  repronim_simple2_brainvolumes.py
#  License: GPL
#**************************************************************************************
#**************************************************************************************
# Date: 03-22-19                 Coded by: David Keator (dbkeator@gmail.com)
# Filename: repronim_simple2_brainvolumes.py
#
# Program description:  This program will load in a CSV file made during simple-2
# brain volumes experiment which has the following organization:
#
#source	FSL	FSL	FSL
# participant_id	left nucleus accumbens volume	left amygdala volume	left caudate nucleus volume
#sub-0050002	796.4723293	1255.574283	4449.579039
#sub-0050003	268.9688215	878.7860634	3838.602449
#sub-0050004	539.0969914	1195.288168	3561.518188
#
# If will use the first row to determine the software used for the segmentations and the
# second row for the variable names.  Then it does a simple NIDM conversion using
# example model in: https://docs.google.com/document/d/1PyBoM7J0TuzTC1TIIFPDqd05nomcCM5Pvst8yCoqLng/edit

#**************************************************************************************
# Development environment: Python - PyCharm IDE
#
#**************************************************************************************
# System requirements:  Python 3.X
# Libraries: pybids, numpy, matplotlib, pandas, scipy, math, dateutil, datetime,argparse,
# os,sys,getopt,csv
#**************************************************************************************
# Start date: 03-22-18
# Update history:
# DATE            MODIFICATION				Who
#
#
#**************************************************************************************
# Programmer comments:
#
#
#**************************************************************************************
#**************************************************************************************

import os,sys
from nidm.experiment import Project,Session,AssessmentAcquisition,AssessmentObject
from nidm.core import Constants
from nidm.experiment.Utils import read_nidm, map_variables_to_terms, getSubjIDColumn
from nidm.experiment.Core import getUUID
from nidm.experiment.Core import Core
from prov.model import QualifiedName,PROV_ROLE, ProvDocument, PROV_ATTR_USED_ENTITY
from prov.model import Namespace as provNamespace
import prov as pm
from argparse import ArgumentParser
from os.path import  dirname, join, splitext,basename
import json
import pandas as pd
from rdflib import Graph,URIRef,RDF
import numpy as np
from io import StringIO
from urllib.parse import urlparse

def column_index(df, query_cols):
    cols = df.columns.values
    sidx = np.argsort(cols)
    return sidx[np.searchsorted(cols,query_cols,sorter=sidx)]

def add_brainvolume_data(nidmdoc, df, id_field, source_row, column_to_terms, png_file=None, output_file=None, root_act=None, nidm_graph=None):
    '''

    :param nidmdoc:
    :param df:
    :param id_field:
    :param source_row:
    :param empty:
    :param png_file:
    :param root_act:
    :return:
    '''
    #dictionary to store activities for each software agent
    software_agent={}
    software_activity={}
    participant_agent={}
    entity={}

    #this function can be used for both creating a brainvolumes NIDM file from scratch or adding brain volumes to
    #existing NIDM file.  The following logic basically determines which route to take...

    #if an existing NIDM graph is passed as a parameter then add to existing file
    if nidm_graph is None:
        first_row=True
        #iterate over rows and store in NIDM file
        for csv_index, csv_row in df.iterrows():

            #store other data from row with columns_to_term mappings
            for row_variable,row_data in csv_row.iteritems():

                #check if row_variable is subject id, if so check whether we have an agent for this participant
                if row_variable==id_field:
                    #store participant id for later use in processing the data for this row
                    participant_id = row_data
                    #if there is no agent for the participant then add one
                    if row_data not in participant_agent.keys():
                        #add an agent for this person
                        participant_agent[row_data] = nidmdoc.graph.agent(QualifiedName(provNamespace("nidm",Constants.NIDM),getUUID()),other_attributes=({Constants.NIDM_SUBJECTID:row_data}))
                    continue
                else:

                    #get source software matching this column deal with duplicate variables in source_row and pandas changing duplicate names
                    software_key = source_row.columns[[column_index(df,row_variable)]]._values[0].split(".")[0]

                    #see if we already have a software_activity for this agent
                    if software_key not in software_activity.keys():

                        #create an activity for the computation...simply a placeholder for more extensive provenance
                        software_activity[software_key] = nidmdoc.graph.activity(QualifiedName(provNamespace("nidm",Constants.NIDM),getUUID()),other_attributes={Constants.NIDM_PROJECT_DESCRIPTION:"brain volume computation"})

                        if root_act is not None:
                            #associate activity with activity of brain volumes creation (root-level activity)
                            software_activity[software_key].add_attributes({QualifiedName(provNamespace("dct",Constants.DCT),'isPartOf'):root_act})

                        #associate this activity with the participant
                        nidmdoc.graph.association(activity=software_activity[software_key],agent=participant_agent[participant_id],other_attributes={PROV_ROLE:Constants.NIDM_PARTICIPANT})
                        nidmdoc.graph.wasAssociatedWith(activity=software_activity[software_key],agent=participant_agent[participant_id])

                        #check if there's an associated software agent and if not, create one
                        if software_key not in software_agent.keys():
                            #create an agent
                            software_agent[software_key] = nidmdoc.graph.agent(QualifiedName(provNamespace("nidm",Constants.NIDM),getUUID()),other_attributes={'prov:type':QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),""),
                                                                    QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),""):software_key } )
                            #create qualified association with brain volume computation activity
                            nidmdoc.graph.association(activity=software_activity[software_key],agent=software_agent[software_key],other_attributes={PROV_ROLE:QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),"")})
                            nidmdoc.graph.wasAssociatedWith(activity=software_activity[software_key],agent=software_agent[software_key])

                    #check if we have an entity for storing this particular variable for this subject and software else create one
                    if software_activity[software_key].identifier.localpart + participant_agent[participant_id].identifier.localpart not in entity.keys():
                        #create an entity to store brain volume data for this participant
                        entity[software_activity[software_key].identifier.localpart + participant_agent[participant_id].identifier.localpart] = nidmdoc.graph.entity( QualifiedName(provNamespace("nidm",Constants.NIDM),getUUID()))
                        #add wasGeneratedBy association to activity
                        nidmdoc.graph.wasGeneratedBy(entity=entity[software_activity[software_key].identifier.localpart + participant_agent[participant_id].identifier.localpart], activity=software_activity[software_key])

                    #get column_to_term mapping uri and add as namespace in NIDM document
                    entity[software_activity[software_key].identifier.localpart + participant_agent[participant_id].identifier.localpart].add_attributes({QualifiedName(provNamespace(Core.safe_string(None,string=str(row_variable)), column_to_terms[row_variable.split(".")[0]]["url"]),""):row_data})
                    #print(project.serializeTurtle())


            #just for debugging.  resulting graph is too big right now for DOT graph creation so here I'm simply creating
            #a DOT graph for the processing of 1 row of the brain volumes CSV file so we can at least visually see the
            #model
            if png_file is not None:
                if first_row:
                    #serialize NIDM file
                    #with open(args.output_file,'w') as f:
                    #   print("Writing NIDM file...")
                    #   f.write(nidmdoc.serializeTurtle())
                    if png_file:
                        nidmdoc.save_DotGraph(str(output_file + ".pdf"), format="pdf")
                    first_row=False
    else:
        first_row=True
        #logic to add to existing graph
        #use RDFLib here for temporary graph making query easier
        rdf_graph = Graph()
        rdf_graph_parse = rdf_graph.parse(source=StringIO(nidmdoc.serializeTurtle()),format='turtle')


        #find subject ids and sessions in NIDM document
        query = """SELECT DISTINCT ?session ?nidm_subj_id ?agent ?entity
                    WHERE {
                        ?activity prov:wasAssociatedWith ?agent ;
                            dct:isPartOf ?session  .
                        ?entity prov:wasGeneratedBy ?activity ;
                            nidm:hadImageUsageType nidm:Anatomical .
                        ?agent rdf:type prov:Agent ;
                            ndar:src_subject_id ?nidm_subj_id .

                    }"""
        #print(query)
        qres = rdf_graph_parse.query(query)



        for row in qres:
            print('%s \t %s' %(row[2],row[1]))
            #find row in CSV file with subject id matching agent from NIDM file

            #csv_row = df.loc[df[id_field]==type(df[id_field][0])(row[1])]
            #find row in CSV file with matching subject id to the agent in the NIDM file
            #be careful about data types...simply type-change dataframe subject id column and query to strings.
            #here we're removing the leading 0's from IDs because pandas.read_csv strips those unless you know ahead of
            #time which column is the subject id....
            csv_row = df.loc[df[id_field].astype('str').str.contains(str(row[1]).lstrip("0"))]

            #if there was data about this subject in the NIDM file already (i.e. an agent already exists with this subject id)
            #then add this brain volumes data to NIDM file, else skip it....
            if (not (len(csv_row.index)==0)):
                print("found other data for participant %s" %row[1])

                #Here we're sure we have an agent in the NIDM graph that corresponds to the participant in the
                #brain volumes data.  We don't know which AcquisitionObject (entity) describes the T1-weighted scans
                #used for the project.  Since we don't have the SHA512 sums in the brain volumes data (YET) we can't
                #really verify that it's a particular T1-weighted scan that was used for the brain volumes but we're
                #simply, for the moment, going to assume it's the activity/session returned by the above query
                #where we've specifically asked for the entity which has a nidm:hasImageUsageType nidm:Anatomical



                #NIDM document entity uuid which has a nidm:hasImageUsageType nidm:Anatomical
                #this is the entity that is associated with the brain volume report for this participant
                anat_entity_uuid = row[3]

                #Now we need to set up the entities/activities, etc. to add the brain volume data for this row of the
                #CSV file and link it to the above entity and the agent for this participant which is row[0]
                #store other data from row with columns_to_term mappings
                for row_variable,row_data in csv_row.iteritems():

                    #check if row_variable is subject id, if so check whether we have an agent for this participant
                    if row_variable==id_field:
                        #store participant id for later use in processing the data for this row
                        participant_id = row_data.values[0]
                        print("participant id: %s" %participant_id)
                        continue
                    else:

                        #get source software matching this column deal with duplicate variables in source_row and pandas changing duplicate names
                        software_key = source_row.columns[[column_index(df,row_variable)]]._values[0].split(".")[0]

                        #see if we already have a software_activity for this agent
                        if software_key+row[2] not in software_activity.keys():

                            #create an activity for the computation...simply a placeholder for more extensive provenance
                            software_activity[software_key+row[2]] = nidmdoc.graph.activity(QualifiedName(provNamespace("niiri",Constants.NIIRI),getUUID()),other_attributes={Constants.NIDM_PROJECT_DESCRIPTION:"brain volume computation",
                                                                                            PROV_ATTR_USED_ENTITY:anat_entity_uuid})

                            #associate the activity with the entity containing the original T1-weighted scan which is stored in anat_entity_uuid
                            if root_act is not None:
                                #associate activity with activity of brain volumes creation (root-level activity)
                                software_activity[software_key+row[2]].add_attributes({QualifiedName(provNamespace("dct",Constants.DCT),'isPartOf'):root_act})



                            #associate this activity with the participant..the participant's agent is row[2] in the query response
                            nidmdoc.graph.association(activity=software_activity[software_key+row[2]],agent=row[2],other_attributes={PROV_ROLE:Constants.NIDM_PARTICIPANT})
                            nidmdoc.graph.wasAssociatedWith(activity=software_activity[software_key+row[2]],agent=row[2])


                            #check if there's an associated software agent and if not, create one
                            if software_key not in software_agent.keys():
                                #if we have a URL defined for this software in Constants.py then use it else simply use the string name of the software product
                                if software_key.lower() in Constants.namespaces:
                                    #create an agent
                                    software_agent[software_key] = nidmdoc.graph.agent(QualifiedName(provNamespace("niiri",Constants.NIIRI),getUUID()),other_attributes={'prov:type':QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),""),
                                                                            QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),""):QualifiedName(provNamespace(software_key,Constants.namespaces[software_key.lower()]),"") } )
                                else:
                                    #create an agent
                                    software_agent[software_key] = nidmdoc.graph.agent(QualifiedName(provNamespace("niiri",Constants.NIIRI),getUUID()),other_attributes={'prov:type':QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),""),
                                                                            QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),""):software_key } )
                            #create qualified association with brain volume computation activity
                            nidmdoc.graph.association(activity=software_activity[software_key+row[2]],agent=software_agent[software_key],other_attributes={PROV_ROLE:QualifiedName(provNamespace(Core.safe_string(None,string=str("Neuroimaging Analysis Software")),Constants.NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE),"")})
                            nidmdoc.graph.wasAssociatedWith(activity=software_activity[software_key+row[2]],agent=software_agent[software_key])

                        #check if we have an entity for storing this particular variable for this subject and software else create one
                        if software_activity[software_key+row[2]].identifier.localpart + row[2] not in entity.keys():
                            #create an entity to store brain volume data for this participant
                            entity[software_activity[software_key+row[2]].identifier.localpart + row[2]] = nidmdoc.graph.entity( QualifiedName(provNamespace("niiri",Constants.NIIRI),getUUID()))
                            #add wasGeneratedBy association to activity
                            nidmdoc.graph.wasGeneratedBy(entity=entity[software_activity[software_key+row[2]].identifier.localpart + row[2]], activity=software_activity[software_key+row[2]])

                        #get column_to_term mapping uri and add as namespace in NIDM document
                        entity[software_activity[software_key+row[2]].identifier.localpart + row[2]].add_attributes({QualifiedName(provNamespace(Core.safe_string(None,string=str(row_variable)), column_to_terms[row_variable.split(".")[0]]["url"]),""):row_data.values[0]})
                        #print(project.serializeTurtle())

            #just for debugging.  resulting graph is too big right now for DOT graph creation so here I'm simply creating
            #a DOT graph for the processing of 1 row of the brain volumes CSV file so we can at least visually see the
            #model
            #if png_file is not None:
            #    if first_row:
                    #serialize NIDM file
                    #with open(args.output_file,'w') as f:
                    #   print("Writing NIDM file...")
                    #   f.write(nidmdoc.serializeTurtle())
            #        nidmdoc.save_DotGraph(str(output_file + ".pdf"), format="pdf")
            #        first_row=False




def main(argv):
    parser = ArgumentParser(description="""This program will load in a CSV file made during simple-2
                brain volumes experiment which has the following organization:
                source	FSL	FSL	FSL
                participant_id	left nucleus accumbens volume	left amygdala volume
                sub-0050002	    796.4723293	    1255.574283	    4449.579039
                sub-0050003	    268.9688215	    878.7860634	    3838.602449
                sub-0050004	    539.0969914	    1195.288168	    3561.518188
                If will use the first row to determine the software used for the segmentations and the
                second row for the variable names.  Then it does a simple NIDM conversion using
                example model in: https://docs.google.com/document/d/1PyBoM7J0TuzTC1TIIFPDqd05nomcCM5Pvst8yCoqLng/edit""")

    parser.add_argument('-csv', dest='csv_file', required=True, help="Path to CSV file to convert")
    parser.add_argument('-ilxkey', dest='key', required=True, help="Interlex/SciCrunch API key to use for query")
    parser.add_argument('-json_map', dest='json_map',required=False,help="User-suppled JSON file containing variable-term mappings.")
    parser.add_argument('-nidm', dest='nidm_file', required=False, help="Optional NIDM file to add CSV->NIDM converted graph to")
    parser.add_argument('-owl', action='store_true', required=False, help='Optionally searches NIDM OWL files...internet connection required')
    parser.add_argument('-png', action='store_true', required=False, help='Optional flag, when set a PNG image file of RDF graph will be produced')
    parser.add_argument('-out', dest='output_file', required=True, help="Filename to save NIDM file")
    args = parser.parse_args()

    #open CSV file and read first line which is the source of the segmentations
    source_row = pd.read_csv(args.csv_file, nrows=0)
    #open CSV file and load into
    df = pd.read_csv(args.csv_file, skiprows=0,header=1)
    #account for duplicate column names
    # df.columns = df.iloc[0]
    df = df.reindex(df.index.drop(0)).reset_index(drop=True)


    #get unique variable names from CSV data file
    #note, duplicate variable names will be appended with a ".X" where X is the number of duplicates
    unique_vars=[]
    for variable in list(df):
        temp=variable.split(".")[0]
        if temp not in unique_vars:
            unique_vars.append(temp)

    #do same as above for unique software agents
    unique_software=[]
    for variable in list(source_row):
        temp=variable.split(".")[0]
        if temp not in unique_software:
            unique_software.append(temp)


    #maps variables in CSV file to terms
    if args.owl:
        column_to_terms = map_variables_to_terms(df=pd.DataFrame(columns=unique_vars), apikey=args.key, directory=dirname(args.output_file), output_file=join(dirname(args.output_file),"json_map.json"), json_file=args.json_map,owl_file=args.owl)
    else:
        column_to_terms = map_variables_to_terms(df=pd.DataFrame(columns=unique_vars), apikey=args.key, directory=dirname(args.output_file), output_file=join(dirname(args.output_file),"json_map.json"), json_file=args.json_map)

    #get subjectID field from CSV
    id_field = getSubjIDColumn(column_to_terms,df)


    # WIP!!!#########################################################################################
    #go line by line through CSV file creating NIDM structures
    #If user has added an existing NIDM file as a command line parameter then add to existing file for subjects who exist in the NIDM file
    if args.nidm_file is not None:
        print("Adding to NIDM file...")
        #read in NIDM file
        project = read_nidm(args.nidm_file)


        root_act = project.graph.activity(QualifiedName(provNamespace("niiri",Constants.NIIRI),getUUID()),other_attributes={Constants.NIDM_PROJECT_DESCRIPTION:"Brain volumes provenance document"})

        #this function sucks...more thought needed for version that works with adding to existing NIDM file versus creating a new NIDM file....
        add_brainvolume_data(nidmdoc=project,df=df,id_field=id_field,root_act=root_act,column_to_terms=column_to_terms,png_file=args.png,output_file=args.output_file,source_row=source_row,nidm_graph=True)

        #serialize NIDM file
        with open(args.output_file,'w') as f:
            print("Writing NIDM file...")
            f.write(project.serializeTurtle())
            #if args.png:
            #    nidmdoc.save_DotGraph(str(args.output_file + ".png"), format="png")



#        #find subject ids and sessions in NIDM document
#        query = """SELECT DISTINCT ?session ?nidm_subj_id ?agent ?entity
#                    WHERE {
#                        ?activity prov:wasAssociatedWith ?agent ;
#                            dct:isPartOf ?session  .
#                        ?entity prov:wasGeneratedBy ?activity ;
#                            nidm:hasImageUsageType nidm:Anatomical .
#                        ?agent rdf:type prov:Agent ;
#                            ndar:src_subject_id ?nidm_subj_id .
#
#                    }"""
#        #print(query)
#        qres = rdf_graph_parse.query(query)



#        for row in qres:
#            print('%s \t %s' %(row[0],row[1]))
#            #find row in CSV file with subject id matching agent from NIDM file

#            #csv_row = df.loc[df[id_field]==type(df[id_field][0])(row[1])]
#            #find row in CSV file with matching subject id to the agent in the NIDM file
#            #be carefull about data types...simply type-change dataframe subject id column and query to strings.
#            #here we're removing the leading 0's from IDs because pandas.read_csv strips those unless you know ahead of
#            #time which column is the subject id....
#            csv_row = df.loc[df[id_field].astype('str').str.contains(str(row[1]).lstrip("0"))]

#            #if there was data about this subject in the NIDM file already (i.e. an agent already exists with this subject id)
#            #then add this brain volumes data to NIDM file, else skip it....
#            if (not (len(csv_row.index)==0)):

                #Here we're sure we have an agent in the NIDM graph that corresponds to the participant in the
                #brain volumes data.  We don't know which AcquisitionObject (entity) describes the T1-weighted scans
                #used for the project.  Since we don't have the SHA512 sums in the brain volumes data (YET) we can't
                #really verify that it's a particular T1-weighted scan that was used for the brain volumes but we're
                #simply, for the moment, going to assume it's the activity/session returned by the above query
                #where we've specifically asked for the entity which has a nidm:hasImageUsageType nidm:Anatomical



                #NIDM document entity uuid which has a nidm:hasImageUsageType nidm:Anatomical
                #this is the entity that is associated with the brain volume report for this participant
#                entity_uuid = row[3]

                #Now we need to set up the entities/activities, etc. to add the brain volume data for this row of the
                #CSV file and link it to the above entity and the agent for this participant which is row[0]





                #add acquisition entity for assessment
#                acq_entity = AssessmentObject(acquisition=acq)
                #add qualified association with existing agent
#                acq.add_qualified_association(person=row[2],role=Constants.NIDM_PARTICIPANT)

#                #store other data from row with columns_to_term mappings
#                for row_variable in csv_row:
                    #check if row_variable is subject id, if so skip it
#                    if row_variable==id_field:
#                        continue
#                    else:
                        #get column_to_term mapping uri and add as namespace in NIDM document
                        #provNamespace(Core.safe_string(None,string=str(row_variable)), column_to_terms[row_variable]["url"])
#                        acq_entity.add_attributes({QualifiedName(provNamespace(Core.safe_string(None,string=str(row_variable)), column_to_terms[row_variable]["url"]), ""):csv_row[row_variable].values[0]})
#                continue

#        #serialize NIDM file
#        with open(args.nidm_file,'w') as f:
#            print("Writing NIDM file...")
#            f.write(project.serializeTurtle())
#            project.save_DotGraph(str(args.nidm_file + ".png"), format="png")
    ##############################################################################################################################


    else:
        print("Creating NIDM file...")
        #If user did not choose to add this data to an existing NIDM file then create a new one for the CSV data

        #create an empty NIDM graph
        nidmdoc = Core()
        root_act = nidmdoc.graph.activity(QualifiedName(provNamespace("niiri",Constants.NIIRI),getUUID()),other_attributes={Constants.NIDM_PROJECT_DESCRIPTION:"Brain volumes provenance document"})

        #this function sucks...more thought needed for version that works with adding to existing NIDM file versus creating a new NIDM file....
        add_brainvolume_data(nidmdoc=nidmdoc,df=df,id_field=id_field,root_act=root_act,column_to_terms=column_to_terms,png_file=args.png,output_file=args.output_file,source_row=source_row)



        #serialize NIDM file
        with open(args.output_file,'w') as f:
            print("Writing NIDM file...")
            f.write(nidmdoc.serializeTurtle())
            if args.png:
            #    nidmdoc.save_DotGraph(str(args.output_file + ".png"), format="png")

                nidmdoc.save_DotGraph(str(args.output_file + ".pdf"), format="pdf")


if __name__ == "__main__":
   main(sys.argv[1:])



#**************************************************************************************
#**************************************************************************************
#  nidm_query.py
#  License: GPL
#**************************************************************************************
#**************************************************************************************
# Date: 8-1-18                 Coded by: David Keator (dbkeator@gmail.com)
# Filename: nidm_query.py
#
# Program description:  This program provides query functionalty for NIDM-Experiment files
#
#
#**************************************************************************************
# Development environment: Python - PyCharm IDE
#
#**************************************************************************************
# System requirements:  Python 3.X
# Libraries: os, sys, rdflib, pandas, argparse, logging
#**************************************************************************************
# Start date: 8-1-18
# Update history:
# DATE            MODIFICATION				Who
#
#
#**************************************************************************************
# Programmer comments:
#
#
#**************************************************************************************
#**************************************************************************************
import os,sys
import rdflib
from rdflib import Graph, URIRef, util
import pandas as pd
import logging
from nidm.core import Constants
import re
import tempfile
from os import path
import functools
import hashlib
from urllib.request import urlretrieve

import pickle

from joblib import Memory
memory = Memory(tempfile.gettempdir(), verbose=0 )


QUERY_CACHE_SIZE = 64
ACQUISITION_MODALITY = 'AcquisitionModality'
IMAGE_CONTRAST_TYPE = 'ImageContrastType'
IMAGE_USAGE_TYPE = 'ImageUsageType'
TASK = 'Task'

def sparql_query_nidm(nidm_file_list,query, output_file=None, return_graph=False):
    '''

    :param nidm_file_list: List of NIDM.ttl files to execute query on
    :param query:  SPARQL query string
    :param output_file:  Optional output file to write results
    :param return_graph: WIP - not working right now but for some queries we prefer to return a graph instead of a dataframe
    :return: dataframe | graph depending on return_graph parameter
    '''

    #query result list
    results = []


    logging.info("Query: %s" , query)

    first_file=True
    #cycle through NIDM files, adding query result to list
    for nidm_file in nidm_file_list:

        # project=read_nidm(nidm_file)
        #read RDF file into temporary graph
        # rdf_graph = Graph()
        # rdf_graph_parse = rdf_graph.parse(nidm_file,format=util.guess_format(nidm_file))
        rdf_graph_parse = OpenGraph(nidm_file)


        if not return_graph:
            #execute query
            qres = rdf_graph_parse.query(query)

            #if this is the first file then grab the SPARQL bound variable names from query result for column headings of query result
            if first_file:
                #format query result as dataframe and return
                #for dicts in qres._get_bindings():
                columns = [str(var) for var in qres.vars]
                first_file=False
                #    break

            #append result as row to result list
            for row in qres:
                results.append(list(row))
        else:
            #execute query
            qres = rdf_graph_parse.query(query)

            if first_file:
                #create graph
                #WIP: qres_graph = Graph().parse(data=qres.serialize(format='turtle'))
                qres_graph = qres.serialize(format='turtle')
                first_file=False
            else:
                #WIP qres_graph = qres_graph + Graph().parse(data=qres.serialize(format='turtle'))
                qres_graph = qres_graph + qres.serialize(format='turtle')



    if not return_graph:
        #convert results list to Pandas DataFrame and return
        df = pd.DataFrame(results,columns=columns)

        #if output file parameter specified
        if (output_file is not None):
            df.to_csv(output_file)
        return df
    else:
        return qres_graph


def GetProjectsUUID(nidm_file_list,output_file=None):
    '''

    :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :return: list of Project UUIDs
    '''

    #SPARQL query to get project UUIDs
    query = '''
        PREFIX nidm:<http://purl.org/nidash/nidm#>
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

        SELECT distinct ?uuid
        Where {
            {
                ?uuid rdf:type nidm:Project
            }

        }
    '''
    df = sparql_query_nidm(nidm_file_list,query, output_file=output_file)

    return df['uuid'].tolist()

def testprojectmeta(nidm_file_list):

    import json

    query = '''
         prefix nidm: <http://purl.org/nidash/nidm#>
         prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

         select distinct ?uuid ?p ?o

         where {
 	        ?uuid rdf:type nidm:Project ;
	   	    ?p  ?o .
         }


    '''

    df =sparql_query_nidm(nidm_file_list,query, output_file=None)

    output_json = {}
    for index,row in df.iterrows():
        if row['uuid'] not in output_json:
            output_json[row['uuid']] = {}

        output_json[row['uuid']][row['p']] = row['o']

    return json.dumps(output_json)

def GetProjectSessionsMetadata(nidm_file_list, project_uuid):

    import json

    query = '''

        prefix nidm: <http://purl.org/nidash/nidm#>
        prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        prefix dct: <http://purl.org/dc/terms/>

        select distinct ?session_uuid ?p ?o

        where {
 	        ?session_uuid  dct:isPartOf  <%s> ;
 	            ?p ?o .
        }

    ''' % project_uuid

    df =sparql_query_nidm(nidm_file_list,query, output_file=None)

    #outermost dictionary
    output_json = {}
    for index,row in df.iterrows():
        if project_uuid not in output_json:
            #creates dictionary for project UUID
            output_json[project_uuid] = {}
        if row['session_uuid'] not in output_json[project_uuid]:
            #creates a dictionary under project_uuid dictionary for session
            output_json[project_uuid][row['session_uuid']] = {}

        output_json[project_uuid][row['session_uuid']][row['p']] = row['o']

    return json.dumps(output_json)


def GetProjectInstruments(nidm_file_list, project_id):
    """
    Returns a list of unique instrument types.  For NIDM files this is rdf:type onli:assessment-instrument
    or related classes (e.g. nidm:NorthAmericanAdultReadingTest, nidm:PositiveAndNegativeSyndromeScale)
    :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :param project_id: identifier of project you'd like to search for unique instruments
    :return: Dataframe of instruments and project titles
    """
    query = '''
        PREFIX prov: <http://www.w3.org/ns/prov#>
        PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
        PREFIX dct: <http://purl.org/dc/terms/>
        prefix onli: <http://neurolog.unice.fr/ontoneurolog/v3.0/instrument.owl#>
        prefix dctypes: <http://purl.org/dc/dcmitype/>

        SELECT  DISTINCT ?project_title ?assessment_type
        WHERE {
            ?entity rdf:type  onli:assessment-instrument ;
                rdf:type ?assessment_type .
            ?entity prov:wasGeneratedBy/dct:isPartOf/dct:isPartOf ?project .

            ?project dctypes:title ?project_title .



            FILTER( (!regex(str(?assessment_type), "http://www.w3.org/ns/prov#Entity")) &&  (!regex(str(?assessment_type), "http://purl.org/nidash/nidm#AcquisitionObject")) &&  (regex(str(?project), "%s")) )
            }
            ''' % project_id
    logging.info('Query: %s', query)
    df = sparql_query_nidm(nidm_file_list, query, output_file=None)
    results = df.to_dict()
    logging.info(results)


    return df

def GetInstrumentVariables(nidm_file_list, project_id):
    '''
    This function will return a comprehensive list of variables as part of any project instrument
    :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :param project_id: identifier of project you'd like to search for unique instruments
    :return: Dataframe of instruments, project titles, and variables
    '''
    query = '''
        PREFIX prov: <http://www.w3.org/ns/prov#>
        PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
        PREFIX dct: <http://purl.org/dc/terms/>
        prefix onli: <http://neurolog.unice.fr/ontoneurolog/v3.0/instrument.owl#>
        prefix dctypes: <http://purl.org/dc/dcmitype/>

        SELECT  DISTINCT ?project_title ?assessment_type ?variables
        WHERE {
            ?entity rdf:type  onli:assessment-instrument ;
                rdf:type ?assessment_type ;
                ?variables ?value .
            ?entity prov:wasGeneratedBy/dct:isPartOf/dct:isPartOf ?project .

            ?project dctypes:title ?project_title .



            FILTER( (!regex(str(?assessment_type), "http://www.w3.org/ns/prov#Entity")) &&  (!regex(str(?assessment_type), "http://purl.org/nidash/nidm#AcquisitionObject")) &&  (regex(str(?project), "%s")) )
            }
            ''' % project_id
    logging.info('Query: %s', query)
    df = sparql_query_nidm(nidm_file_list, query, output_file=None)
    results = df.to_dict()
    logging.info(results)


    return df

def GetParticipantIDs(nidm_file_list,output_file=None):
    '''
    This query will return a list of all prov:agent entity UUIDs that prov:hadRole sio:Subject or Constants.NIDM_PARTICIPANT
    :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :return: list of Constants.NIDM_PARTICIPANT UUIDs and Constants.NIDM_SUBJECTID
    '''

    query = '''

        PREFIX prov:<http://www.w3.org/ns/prov#>
        PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
        PREFIX ndar: <https://ndar.nih.gov/api/datadictionary/v2/dataelement/>
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX prov:<http://www.w3.org/ns/prov#>

        SELECT DISTINCT ?uuid ?ID
        WHERE {

            ?activity rdf:type prov:Activity ;
		        prov:qualifiedAssociation _:blanknode .

	        _:blanknode prov:hadRole %s ;
                 prov:agent ?uuid  .

	        ?uuid %s ?ID .

        }
    ''' %(Constants.NIDM_PARTICIPANT,Constants.NIDM_SUBJECTID)

    df = sparql_query_nidm(nidm_file_list,query, output_file=output_file)

    return df


def GetParticipantDetails(nidm_file_list,project_id, participant_id, output_file=None):
    '''
    This query will return a list of all prov:agent entity UUIDs that prov:hadRole Constants.NIDM_PARTICIPANT
    :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :return: list of Constants.NIDM_PARTICIPANT UUIDs and Constants.NIDM_SUBJECTID
    '''

    query = '''

        PREFIX prov:<http://www.w3.org/ns/prov#>
        PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
        PREFIX ndar: <https://ndar.nih.gov/api/datadictionary/v2/dataelement/>
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX nidm: <http://purl.org/nidash/nidm#>
        PREFIX dct: <http://purl.org/dc/terms/>


        SELECT DISTINCT ?uuid ?id ?activity
        WHERE {

            ?activity rdf:type prov:Activity ;
		        prov:qualifiedAssociation _:blanknode .

	        _:blanknode prov:hadRole %s ;
                 prov:agent ?uuid  .

	        ?uuid %s ?id .

            ?proj a nidm:Project .
            ?sess dct:isPartOf ?proj .
            ?activity dct:isPartOf ?sess .

            FILTER(regex(str(?uuid), "%s")).

        }
    ''' %(Constants.NIDM_PARTICIPANT,Constants.NIDM_SUBJECTID, participant_id)

    df = sparql_query_nidm(nidm_file_list,query, output_file=output_file)
    data = df.values

    uuid = ""
    id = ""
    if len(data) > 0:
        uuid = data[0][0]
        id = data[0][1]

    result = { 'uuid' : str(uuid).replace(Constants.NIIRI, ""),
               'id' : str(id),
               'activity': [] }

    for row in data:
        act = (str(row[2])).replace(str(Constants.NIIRI), "")
        (result['activity']).append( act )

    result["instruments"] = GetParticipantInstrumentData(nidm_file_list, project_id, participant_id)

    result["derivatives"] = GetDerivativesDataForSubject(nidm_file_list, None, participant_id)

    return result

def GetMergedGraph(nidm_file_list):
    rdf_graph = Graph()
    for f in nidm_file_list:
        rdf_graph.parse(f, format=util.guess_format(f))
    return rdf_graph

def GetNameForDataElement(graph, uri):
    label = isAbout = source_variable = None


    for data_element, predicate, value in graph.triples( (uri, None, None) ):
        if predicate == Constants.NIDM['source_variable']:
            source_variable = str(value)
        if predicate == Constants.NIDM['isAbout']:
            isAbout = str(value)
        if predicate == Constants.RDFS['label']:
            label = str(value)

    return source_variable or label or isAbout or URITail(uri)


def GetParticipantInstrumentData(nidm_file_list ,project_id, participant_id):
    return GetParticipantInstrumentDataCached(tuple(nidm_file_list) ,project_id, participant_id)

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def GetParticipantInstrumentDataCached(nidm_file_list: tuple ,project_id, participant_id):
    '''
    This query will return a list of all instrument data for prov:agent entity UUIDs that has
    prov:hadRole sio:Subject or Constants.NIDM_PARTICIPANT
    :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :return: list of Constants.NIDM_PARTICIPANT UUIDs and Constants.NIDM_SUBJECTID
    '''

    if participant_id.find('http') != 0:
        participant_id = Constants.NIIRI[participant_id]

    result = {}
    names = []
    for f in nidm_file_list:
        rdf_graph = OpenGraph(f)
        for n in rdf_graph.namespace_manager.namespaces():
            if not n in names:
                names.append(n)

    isa = URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
    for f in nidm_file_list:
        rdf_graph = OpenGraph(f)
        # find all the instrument based assessments
        for acquisition in rdf_graph.subjects(isa, Constants.NIDM['Acquisition']):
            # verify that the assessment is linked to a subject through a blank node
            for blanknode in rdf_graph.objects(subject=acquisition,predicate=Constants.PROV['qualifiedAssociation']):
                # check to see if this assessment is about our participant
                if ((blanknode, Constants.PROV['agent'], participant_id) in rdf_graph)  :
                    # now we know that the assessment is one we want, find the actual assessment data
                    for instrument in rdf_graph.subjects(predicate=Constants.PROV['wasGeneratedBy'], object=acquisition):
                        #load up all the assement data into the result
                        instrument_key = str(instrument).split('/')[-1]
                        result[instrument_key] = {}
                        for s,data_element,o in rdf_graph.triples((instrument, None, None)):
                            # convert the random looking URIs to the prefix used in the ttl file, if any
                            matches = [n[0] for n in names if n[1] == data_element]
                            if len(matches) > 0:
                                idx = str(matches[0])
                            else:
                                # idx = str(data_element)
                                idx = GetNameForDataElement(rdf_graph, data_element)
                            result[instrument_key][ idx ] = str(str(o))


    return result

def GetParticipantUUIDsForProject(nidm_file_list: tuple, project_id, filter, output_file=None):
    return GetParticipantUUIDsForProjectCached(tuple(nidm_file_list), project_id, filter, output_file=None)

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def GetParticipantUUIDsForProjectCached(nidm_file_list:tuple, project_id, filter, output_file=None):
    '''
    This query will return a list of all prov:agent entity UUIDs within a single project
    that prov:hadRole sio:Subject or Constants.NIDM_PARTICIPANT
    :param filter:
    :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :return: list of Constants.NIDM_PARTICIPANT UUIDs and Constants.NIDM_SUBJECTID
    '''

    # if this isn't already a URI, make it one.
    # calls from the REST api don't include the URI
    project = project_id
    if project_id.find('http') < 0:
        project = Constants.NIIRI[project_id]


    participants = []


    for file in nidm_file_list:
        rdf_graph = OpenGraph(file)
        #find all the sessions
        for (session, p, o) in rdf_graph.triples((None, None, Constants.NIDM['Session'])): #rdf_graph.subjects(object=isa, predicate=Constants.NIDM['Session']):
            #check if it is part of our project
            if (session, Constants.DCT['isPartOf'], project) in rdf_graph:
                #find all the activities/acquisitions/etc that are part of this session
                for activity in rdf_graph.subjects(predicate=Constants.DCT['isPartOf'], object=session):
                    # look to see if the activity is linked to a subject via blank node
                    for blank in rdf_graph.objects(subject=activity, predicate=Constants.PROV['qualifiedAssociation']):
                        if (blank, Constants.PROV['hadRole'], Constants.SIO['Subject']):
                            for participant in rdf_graph.objects(subject=blank, predicate=Constants.PROV['agent']):
                                uuid = (str(participant)).split('/')[-1]  # srip off the http://whatever/whatever/
                                if (not uuid in participants) and \
                                        ( (not filter) or CheckSubjectMatchesFilter( tuple([file]) , project, participant, filter) ):
                                    participants.append(uuid)

    return participants


# if this isn't already a URI, make it one.
# calls from the REST api don't include the URI
def expandUUID(partial_uuid):
    uuid = partial_uuid
    if partial_uuid.find('http') < 0:
        uuid = Constants.NIIRI[partial_uuid]
    return uuid


def getProjectAcquisitionObjects(nidm_file_list, project_id):
    acq_objects = []
    isa = URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
    project_uuid = expandUUID(project_id)

    for file in nidm_file_list:
        rdf_graph = OpenGraph(file)
        #find all the projects
        for (project,pred,o) in rdf_graph.triples((None, None, Constants.NIDM['Project'])):
            #check if it is our project
            if str(project) == project_uuid:
                for (session,p2,o2) in rdf_graph.triples((None,isa, Constants.NIDM['Session'])):
                    for (acquisition,p3,o3) in rdf_graph.triples((None, Constants.DCT['isPartOf'], session)):
                        for (acq_obj, p4, o4) in rdf_graph.triples((None, Constants.PROV['wasGeneratedBy'], acquisition)):
                            if (acq_obj, isa, Constants.NIDM['AcquisitionObject']):
                                acq_objects.append(acq_obj)
    return acq_objects


def GetProjectDataElements(nidm_file_list, project_id):
    result = []
    isa = URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')

    # if this isn't already a URI, make it one.
    # calls from the REST api don't include the URI
    project = project_id
    if project_id.find('http') < 0:
        project = Constants.NIIRI[project_id]

    for file in nidm_file_list:
        rdf_graph = OpenGraph(file)
        #find all the sessions
        for (session, p, o) in rdf_graph.triples((None, None, Constants.NIDM['Session'])): #rdf_graph.subjects(object=isa, predicate=Constants.NIDM['Session']):
            #check if it is part of our project
            if (session, Constants.DCT['isPartOf'], project) in rdf_graph:
                # we know we have the right file, so just grab all the data elements from here
                for de in rdf_graph.subjects(isa, Constants.NIDM['DataElement']):
                    result.append(rdf_graph.namespace_manager.compute_qname(str(de))[0])
                return result
    return result


# in case someone passes in a filter subject with a full http or https URI, strip it back to just the bit after the namespace
def splitSubject(subject):
    if subject.find("http") > -1:
        matches = re.match(r'.*(https?://[^/]+[^\. ]+)', subject)
        URI = matches.group(1)
        subject = str(subject).replace(URI, URITail(URI))

    return subject.split(".")

def URITail(URI):
    '''
    Returns the last bit of a URI.
    Useful for pulling out datatype from long namespaces , e.g. http://purl.org/nidash/fsl#fsl_000032
    :param URI: string
    :return: string
    '''
    tail = URI.split('/')[-1]
    tail = tail.split('#')[-1]
    return  tail

def trimWellKnownURIPrefix(uri):
    trimed = uri
    for p in ['http://purl.org/nidash/nidm#', 'http://www.w3.org/ns/prov#', 'http://iri.nidash.org/']:
        trimed = str(trimed).replace(p, '')
    return trimed

def CheckSubjectMatchesFilter(nidm_file_list, project_uuid, subject_uuid, filter):
    '''
    filter should look something like:
       projects.subjects.instruments.AGE gt 12 and projects.subjects.instruments.SITE_ID eq CMU

    :param nidm_file_list:
    :param project_uuid:
    :param subject_uuid:
    :param filter:
    :return:
    '''

    # TODO: I need to fix this here.  When there is a space inside the value the splitter gets more than 3 values
    # ex: 'projects.subjects.instruments.WISC_IV_VOCAB_SCALED eq \'not a match\''

    # filter can have multiple and clauses, break them up and test each one
    tests = filter.split('and')



    for test in tests:
        found_match = False
        split_array = test.split(' ')
        compound_sub = split_array[0]
        op = split_array[1]
        value = ' '.join(split_array[2:])

        #if the value is a string, it will have quotes around it.  Strip them out now
        for quote in ["'", "\"", "`"]:
            if value[0] == quote and value[-1] == quote:
                value = value[1:-1]


        sub_pieces = splitSubject(compound_sub)
        if len(sub_pieces) == 2 and sub_pieces[0] == 'instruments':
            term = sub_pieces[1] # 'AGE_AT_SCAN' for example
            instrument_details = GetParticipantInstrumentData(nidm_file_list, project_uuid, subject_uuid)
            for instrument_uuid in instrument_details:
                for instrument_term in instrument_details[instrument_uuid]:
                    if instrument_term == term:
                        found_match = filterCompare(instrument_details[instrument_uuid][term], op, value)
                    if found_match:
                        break

        elif len(sub_pieces) == 2 and sub_pieces[0] == 'derivatives':
            type = sub_pieces[1] # 'ilx:0102597' for example
            derivatives_details = GetDerivativesDataForSubject(nidm_file_list, project_uuid, subject_uuid)
            for key in derivatives_details:
                derivatives = derivatives_details[key]['values']
                for vkey in derivatives:  # values will be in the form { http://example.com/a/b/c#fs_00001 : { datumType: '', label: '', value: '', units:'' }, ... }
                    short_key = URITail(vkey)
                    if short_key == type:
                        found_match = filterCompare(derivatives[vkey]['value'], op, value)
                    if found_match:
                        break




        # check after each test if we got false because the tests are joined with 'and'
        if not found_match:
            return False

    return True

def filterCompare(left, op, right):
    try:
        if op == 'eq':
            return (left == right)
        elif op == 'lt':
            return (float(left) < float(right))
        elif op == 'gt':
            return (float(left) > float(right))
    except:
        pass

    return None

def GetProjectsMetadata(nidm_file_list):
    '''
     :param nidm_file_list: List of one or more NIDM files to query for project meta data
    :return: dataframe with two columns: "project_uuid" and "project_dentifier"
    '''

    query = '''
        PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX nidm:<http://purl.org/nidash/nidm#>
         SELECT DISTINCT ?property ?o ?s WHERE {{ ?s a nidm:Project . ?s ?property ?o }}
    '''
    df = sparql_query_nidm(nidm_file_list, query, output_file=None)

    projects = {}
    arr = df.values

    for row in arr:
        field = str(row[0])
        value = str(row[1])
        project = str(row[2])
        if project not in projects:
            projects[str(project)] = {}
        # if field in field_whitelist:
        projects[str(project)][field] = value

    return {'projects': compressForJSONResponse(projects)}


def GetProjectsComputedMetadata(nidm_file_list):
    '''
     :param nidm_file_list: List of one or more NIDM files to query across for list of Projects
    :return: dataframe with two columns: "project_uuid" and "project_dentifier"
    '''

    meta_data = GetProjectsMetadata(nidm_file_list)
    ExtractProjectSummary(meta_data, nidm_file_list)

    return compressForJSONResponse(meta_data)

def GetDataElements(nidm_file_list):

    query='''
        select distinct ?uuid ?DataElements
            where {

                ?uuid a ?DataElements

                filter( regex(str(?DataElements), "DataElement" ))

            }'''

    df = sparql_query_nidm(nidm_file_list.split(','), query, output_file=None)
    return df
def GetBrainVolumeDataElements(nidm_file_list):
    query='''
        prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        prefix prov: <http://www.w3.org/ns/prov#>
        prefix ndar: <https://ndar.nih.gov/api/datadictionary/v2/dataelement/>
        prefix fsl: <http://purl.org/nidash/fsl#>
        prefix nidm: <http://purl.org/nidash/nidm#>
        prefix onli: <http://neurolog.unice.fr/ontoneurolog/v3.0/instrument.owl#>
        prefix freesurfer: <https://surfer.nmr.mgh.harvard.edu/>
        prefix dx: <http://ncitt.ncit.nih.gov/Diagnosis>
        prefix ants: <http://stnava.github.io/ANTs/>
        prefix dct: <http://purl.org/dc/terms/>
        prefix dctypes: <http://purl.org/dc/dcmitype/>

        SELECT DISTINCT ?tool ?softwareLabel ?federatedLabel ?laterality
        where {
 	        ?tool_act a prov:Activity ;
		            prov:qualifiedAssociation [prov:agent [nidm:NIDM_0000164 ?tool]] .
			?tool_entity prov:wasGeneratedBy ?tool_act ;
				?measure ?volume .

			{?measure a fsl:DataElement ;
				    rdfs:label ?softwareLabel;
				    nidm:measureOf <http://uri.interlex.org/base/ilx_0112559> ;
				    nidm:datumType <http://uri.interlex.org/base/ilx_0738276> ;
			}
			UNION
			{?measure a freesurfer:DataElement ;
				    rdfs:label ?softwareLabel;
				    nidm:measureOf <http://uri.interlex.org/base/ilx_0112559> ;
				    nidm:datumType <http://uri.interlex.org/base/ilx_0738276> ;
			}
			UNION
			{?measure a ants:DataElement ;
				    rdfs:label ?softwareLabel;
				    nidm:measureOf <http://uri.interlex.org/base/ilx_0112559> ;
				    nidm:datumType <http://uri.interlex.org/base/ilx_0738276> ;
			}
			OPTIONAL {?measure nidm:isAbout ?federatedLabel }.
			OPTIONAL {?measure nidm:hasLaterality ?laterality }.
		}'''

    df = sparql_query_nidm(nidm_file_list.split(','), query, output_file=None)
    return df

def GetBrainVolumes(nidm_file_list):
    query='''
        # This query simply returns the brain volume data without dependencies on other demographics/assessment measures.

        prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        prefix prov: <http://www.w3.org/ns/prov#>
        prefix ndar: <https://ndar.nih.gov/api/datadictionary/v2/dataelement/>
        prefix fsl: <http://purl.org/nidash/fsl#>
        prefix nidm: <http://purl.org/nidash/nidm#>
        prefix onli: <http://neurolog.unice.fr/ontoneurolog/v3.0/instrument.owl#>
        prefix freesurfer: <https://surfer.nmr.mgh.harvard.edu/>
        prefix dx: <http://ncitt.ncit.nih.gov/Diagnosis>
        prefix ants: <http://stnava.github.io/ANTs/>
        prefix dct: <http://purl.org/dc/terms/>
        prefix dctypes: <http://purl.org/dc/dcmitype/>
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        select distinct ?ID ?tool ?softwareLabel ?federatedLabel ?laterality ?volume
        where {
 	        ?tool_act a prov:Activity ;
		            prov:qualifiedAssociation [prov:agent [nidm:NIDM_0000164 ?tool]] .
			?tool_act prov:qualifiedAssociation [prov:agent [ndar:src_subject_id ?ID]] .
			?tool_entity prov:wasGeneratedBy ?tool_act ;
				?measure ?volume .

				?tool_entity prov:wasGeneratedBy ?tool_act ;
					?measure ?volume .

					?measure a/rdfs:subClassOf* nidm:DataElement ;
						 rdfs:label ?softwareLabel;
						 nidm:measureOf <http://uri.interlex.org/base/ilx_0112559> ;
						 nidm:datumType <http://uri.interlex.org/base/ilx_0738276> .
					OPTIONAL {?measure nidm:isAbout ?federatedLabel }.
					OPTIONAL {?measure nidm:hasLaterality ?laterality }.

            }
            '''

    df = sparql_query_nidm(nidm_file_list.split(','), query, output_file=None)
    return df



def ExtractProjectSummary(meta_data, nidm_file_list):
    '''

    :param meta_data: a dictionary of projects containing their meta data as pulled from the nidm_file_list
    :param nidm_file_list: List of NIDM files
    :return:
    '''
    query = '''
    prefix ncicb: <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#>
    prefix ndar: <https://ndar.nih.gov/api/datadictionary/v2/dataelement/>
    prefix obo: <http://purl.obolibrary.org/obo/>
    prefix dct: <http://purl.org/dc/terms/>
    prefix prov: <http://www.w3.org/ns/prov#>
    prefix nidm: <http://purl.org/nidash/nidm#>

    SELECT DISTINCT ?id ?person ?age ?gender ?hand ?assessment ?acq ?session ?project
    WHERE {
      # Added by DBK to support new data element format
      {?age_measure a nidm:DataElement ;
					nidm:isAbout ncicb:Age .}
	  {?gender_measure a nidm:DataElement ;
					nidm:isAbout ndar:gender .}
	  {?handedness_measure a nidm:DataElement ;
					nidm:isAbout obo:handedness .}
      OPTIONAL { ?assessment ?age_measure ?age } .
      OPTIONAL { ?assessment ?gender_measure ?gender } .
      OPTIONAL { ?assessment ?handedness_measure ?hand } .
      ?person ndar:src_subject_id ?id .
      ?acq prov:qualifiedAssociation _:blank .
      _:blank prov:hadRole sio:Subject .
      _:blank prov:agent ?person .
      ?assessment prov:wasGeneratedBy ?acq .
      ?acq dct:isPartOf ?session .
      ?session dct:isPartOf ?project .
      ?project a nidm:Project
    }
    ORDER BY ?id
      '''

    df = sparql_query_nidm(nidm_file_list, query, output_file=None)
    projects = meta_data['projects']

    arr = df.values
    key = str(Constants.NIDM_NUMBER_OF_SUBJECTS)
    for project_id, project in projects.items():
        project[key] = 0
        project['age_max'] = 0
        project['age_min'] = sys.maxsize
        project[str(Constants.NIDM_GENDER)] = []
        project[str(Constants.NIDM_HANDEDNESS)] = []

    for row in arr:
        project_id = matchPrefix( str(row[8]) ) # 9th column is the project UUID
        projects[project_id][str(Constants.NIDM_NUMBER_OF_SUBJECTS)] += 1

        age = float(row[2])
        projects[project_id]['age_min'] = min(age, projects[project_id]['age_min'])
        projects[project_id]['age_max'] = max(age, projects[project_id]['age_max'])

        gender = str(row[3])
        if gender not in projects[project_id][str(Constants.NIDM_GENDER)]:
            projects[project_id][str(Constants.NIDM_GENDER)].append(gender)

        hand = str(row[4])
        if hand not in projects[project_id][str(Constants.NIDM_HANDEDNESS)]:
            projects[project_id][str(Constants.NIDM_HANDEDNESS)].append(hand)


def expandNIDMAbbreviation(shortKey) -> str:
    '''
    Takes a shorthand identifier such as dct:description and returns the 
    full URI http://purl.org/dc/terms/description

    :param shortKey:
    :type shortKey: str
    :return:
    '''
    newkey = skey = str(shortKey)
    match = re.search(r"^([^:]+):([^:]+)$", skey)
    if match:
        newkey = Constants.namespaces[match.group(1)] + match.group(2)
    return newkey

def compressForJSONResponse(data) -> dict:
    '''
    Takes a Dictionary and shortens any key by replacing a full URI with
    the NIDM prefix

    :param data: Data to seach for long URIs that can be replaced with prefixes
    :return: Dictionary
    '''
    new_dict = {}

    if isinstance(data,dict):
        for key, value in data.items():
            new_dict[matchPrefix(key)] = compressForJSONResponse(value)
    else:
        return data

    return new_dict

def matchPrefix(possible_URI, short=False) -> str:
    '''
    If the possible_URI is found in Constants.namespaces it will
    be replaced with the prefix

    :param possible_URI: URI string to look at
    :type possible_URI: str
    :return: Returns a
    '''
    for k, n in Constants.namespaces.items():
        if possible_URI.startswith(n):
            if short:
                return k
            else:
                return "{}:{}".format(k, possible_URI.replace(n, ""))

    # also check the prov prefix
    if possible_URI.startswith("http://www.w3.org/ns/prov#"):
        return "{}:{}".format("prov", possible_URI.replace("http://www.w3.org/ns/prov#", ""))

    return possible_URI

# check if this activity is linked by a blank node to one of the sw agents
def activityIsSWAgent(rdf_graph, activity, sw_agents):
    '''
    Returns True if the given activity is associated with a software agent from the sw_agents array
    :param rdf_graph: Graph
    :param activity: activity URI
    :param sw_agents: array of software agent URIs
    :return: Boolean
    '''
    if activity in sw_agents:
        return True

    return False

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def getDerivativesNodesForSubject (rdf_graph, subject):
    '''
    Finds all the URIs that were generated by software agents and linked to the subject

    :param rdf_graph:
    :param subject:
    :return: Array of StatsCollections URIs
    '''

    qualified_association = URIRef('http://www.w3.org/ns/prov#qualifiedAssociation')
    was_associated_with = URIRef('http://www.w3.org/ns/prov#wasAssociatedWith')

    sw_agents = getSoftwareAgents(rdf_graph)
    derivatives_uris = []

    for blank, p, o in rdf_graph.triples( (None, Constants.PROV['agent'], subject)): # get the blank nodes associated with the subject
        # verify this blank node points to the subject somewhere it has the  role subject
        if (blank, Constants.PROV['hadRole'], Constants.SIO['Subject']) in rdf_graph:
            # get the activity that's the parent of the blank node
            for activity in rdf_graph.subjects(predicate=Constants.PROV['qualifiedAssociation'], object=blank):
                # try to find if this activity has a qualified association with a software agent (through a blank node)
                for software_blank in rdf_graph.objects(subject=activity, predicate=Constants.PROV['qualifiedAssociation']):
                    for software_agent in rdf_graph.objects(subject=software_blank, predicate=Constants.PROV['agent']):
                        if activityIsSWAgent(rdf_graph, software_agent, sw_agents):
                            # now we know our activity generated a stats colleciton, so go find it (the stats_colleciton will be generated by the activity)
                            for stats_collection in rdf_graph.subjects(predicate=Constants.PROV['wasGeneratedBy'], object=activity):
                                derivatives_uris.append(stats_collection)

    return derivatives_uris

def getDataTypeInfo(source_graph, datatype):
    '''
    Scans all the triples with subject of datatype (isa DataElement in the graph) and looks for entries
    with specific predicates necessary to define it's type

    :param rdf_graph:
    :param datatype: URI of the DataElement
    :return: { 'label': label, 'hasUnit': hasUnit, 'typeURI': typeURI}
    '''

    # check to see if the datatype is in the main graph. If not, look in the CDE graph
    if (datatype, None, None) in source_graph:
        rdf_graph = source_graph
    else:
        rdf_graph = getCDEs()

    typeURI = ''
    hasUnit = ''
    label = ''
    desc = ''
    measureOf = ''
    isAbout = ''
    structure = ''



    # have to scan all tripples because the label can be in any namespace
    for s, p, o in rdf_graph.triples((datatype, None, None)):
        if (re.search(r'label$', str(p)) != None):
            label = o
        if (re.search(r'hasUnit$', str(p), flags=re.IGNORECASE) != None):
            hasUnit = o
        if (re.search(r'datumType$', str(p)) != None):
            typeURI = str(o).split('/')[-1]
        if (re.search(r'measureOf$', str(p)) != None):
            measureOf = o
        if (re.search(r'isAbout$', str(p), flags=re.IGNORECASE) != None):
            isAbout = o

    return {'label': label, 'hasUnit': hasUnit, 'datumType': typeURI, 'measureOf': measureOf, 'isAbout': isAbout}

def getStatsCollectionForNode (rdf_graph, derivatives_node):

    isa = URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
    data = {'URI': derivatives_node, 'values': {}}

    for s, datatype, value in rdf_graph.triples((derivatives_node, None, None)):
        if datatype == isa and str(value).find('http://purl.org/nidash/nidm#') == 0:
            data['StatCollectionType'] = str(value)[28:]
        else:
            dti = getDataTypeInfo(rdf_graph, datatype )
            data['values'][str(datatype)] = {'datumType': str(dti['datumType']), 'label': str(dti['label']), 'value': str(value), 'units': str(dti['hasUnit'])}

    return data

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def OpenGraph(file):
    '''
    Returns a parsed RDFLib Graph object for the given file
    The file will be hashed and if a pickled copy is found in the TMP dir, that will be used
    Otherwise the graph will be computed and then saved in the TMP dir as a pickle file
    We also use functools.lru_cache to cache results in memory during a run

    :param file: filename
    :return: Graph
    '''
    # if someone passed me a RDF graph rather than a file, just send it back
    if isinstance(file, rdflib.graph.Graph):
        return file

    BLOCKSIZE = 65536
    hasher = hashlib.md5()
    with open(file, 'rb') as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)
    hash = hasher.hexdigest()

    pickle_file = '{}/rdf_graph.{}.pickle'.format( tempfile.gettempdir(), hash)
    if path.isfile(pickle_file):
        return pickle.load(open(pickle_file, "rb"))

    rdf_graph = Graph()
    rdf_graph.parse(file, format=util.guess_format(file))
    pickle.dump(rdf_graph, open(pickle_file, 'wb'))

    # new graph, so to be safe clear out all cached entries
    memory.clear(warn=False)

    return rdf_graph

def GetDerivativesDataForSubject(files, project, subject):
    return GetDerivativesDataForSubjectCache (tuple(files), project, subject)

@functools.lru_cache(maxsize=QUERY_CACHE_SIZE)
def GetDerivativesDataForSubjectCache(files, project, subject):
    '''
    Searches for the subject in the supplied RDF .ttl files and returns
    an array of all the data generated by software agents about that subject

    :param project:
    :param files: Array of RDF .ttl files
    :param subject: The URI (or just the bit after the NIIRI prefix) of a subject
    :return: Array of stat collections for the subject
    '''

    # if this isn't already a URI, make it one.
    # calls from the REST api don't include the URI
    if subject.find('http') < 0:
        subject = Constants.NIIRI[subject]

    data = {}

    for nidm_file in files:
        rdf_graph = OpenGraph(nidm_file)
        for node in getDerivativesNodesForSubject(rdf_graph, subject):
            collection = getStatsCollectionForNode(rdf_graph, node)
            key = str(collection['URI']).split('/')[-1]
            data[key] = collection

    return data

def getSoftwareAgents(rdf_graph):
    '''
    Scans the supplied graph and returns any software agenyt URIs found there

    :param rdf_graph: a parsed RDF Graph
    :return: array of agent URIs
    '''

    isa = URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
    software_agent = URIRef('http://www.w3.org/ns/prov#SoftwareAgent')
    agents = []

    for s,o,p in rdf_graph.triples( (None, isa, software_agent) ):
        agents.append(s)

    return agents

def download_cde_files():
    cde_dir = tempfile.gettempdir()

    for url in Constants.CDE_FILE_LOCATIONS:
        urlretrieve( url, "{}/{}".format(cde_dir, url.split('/')[-1] ) )

    return cde_dir


def getCDEs(file_list=None):

    if getCDEs.cache:
        return getCDEs.cache

    hasher = hashlib.md5()
    hasher.update(str(file_list).encode('utf-8'))
    h = hasher.hexdigest()

    cache_file_name = tempfile.gettempdir() + "/cde_graph.{}.pickle".format(h)

    if path.isfile(cache_file_name):
        rdf_graph = pickle.load(open(cache_file_name, "rb"))
        getCDEs.cache = rdf_graph
        return rdf_graph

    rdf_graph = Graph()

    if not file_list:

        cde_dir = ''
        if "CDE_DIR" in os.environ:
            cde_dir = os.environ['CDE_DIR']

        if (not cde_dir) and (os.path.isfile( '/opt/project/nidm/core/cde_dir/ants_cde.ttl' )):
            cde_dir = '/opt/project/nidm/core/cde_dir'

        if (not cde_dir):
            cde_dir = download_cde_files()

        file_list = [ ]
        for f in ['ants_cde.ttl', 'fs_cde.ttl', 'fsl_cde.ttl']:
            fname = '{}/{}'.format(cde_dir, f)
            if os.path.isfile( fname ):
                file_list.append( fname )



    for fname in file_list:
        if os.path.isfile(fname):
            cde_graph = OpenGraph(fname)
            rdf_graph = rdf_graph + cde_graph




    cache_file = open(cache_file_name , 'wb')
    pickle.dump(rdf_graph, cache_file)
    cache_file.close()

    getCDEs.cache = rdf_graph
    return rdf_graph
getCDEs.cache = None